# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
#
################################################################################
# Utility functions for Ray-tracing
################################################################################
cimport cython
from libc.math cimport sqrt as Csqrt, fabs as Cabs
from libc.math cimport cos as Ccos, sin as Csin
from libc.math cimport atan2 as Catan2
from libc.math cimport NAN as Cnan
from libc.math cimport pi as Cpi
from libc.math cimport isnan as Cisnan
from cython.parallel import prange
from cython.parallel cimport parallel
from cpython.array cimport array, clone
from libc.stdlib cimport malloc, free
cimport numpy as np
import numpy as np
from cython cimport view
# importing ToFu functions:
from _basic_geom_tools cimport _VSMALL
from _basic_geom_tools cimport is_point_in_path
from _basic_geom_tools cimport is_point_in_path_vec
from _basic_geom_tools cimport compute_inv_and_sign
cimport _basic_geom_tools as _bgt

# ==============================================================================
# =  3D Bounding box (not Toroidal)
# ==============================================================================
cdef inline void compute_3d_bboxes(const double*const* vignett_poly,
                                   const long* lnvert,
                                   const int nvign,
                                   double* lbounds,
                                   const int num_threads) nogil:
    """
    Computes coordinates of bounding boxes of a list of 3d objects in a general
    space (not related to a tore).
    """
    cdef int ivign
    cdef int nvert
    # ...
    # -- Defining parallel part ------------------------------------------------
    with nogil, parallel(num_threads=num_threads):
        for ivign in prange(nvign):
            nvert = lnvert[ivign]
            comp_bbox_poly3d(nvert,
                             &vignett_poly[ivign][0*nvert],
                             &vignett_poly[ivign][1*nvert],
                             &vignett_poly[ivign][2*nvert],
                             &lbounds[ivign*6])
    return

cdef inline void comp_bbox_poly3d(const int nvert,
                                  const double* vertx,
                                  const double* verty,
                                  const double* vertz,
                                  double[6] bounds) nogil:
    """
    Computes bounding box of a 3d polygon
    Params
    =====
    nvert : integer
       Number of vertices in the poygon
    vert : double array
       Coordinates of the polygon defining the structure in the poloidal plane
       such that vert[0:3, ii] = (x_i, y_i, z_i) the coordinates of the i-th
       vertex
    bounds : (6) double array <INOUT>
       coordinates of the lowerleftback point and of the upperrightfront point
       of the bounding box of the polygon
    """
    cdef int ii
    cdef double xmax=vertx[0], xmin=vertx[0]
    cdef double ymax=verty[0], ymin=verty[0]
    cdef double zmax=vertz[0], zmin=vertz[0]
    cdef double tmp_val
    for ii in range(1, nvert):
        # x....
        tmp_val = vertx[ii]
        if tmp_val > xmax:
            xmax = tmp_val
        elif tmp_val < xmin :
            xmin = tmp_val
        # y....
        tmp_val = verty[ii]
        if tmp_val > ymax:
            ymax = tmp_val
        elif tmp_val < ymin :
            ymin = tmp_val
        # z....
        tmp_val = vertz[ii]
        if tmp_val > zmax:
            zmax = tmp_val
        elif tmp_val < zmin :
            zmin = tmp_val
    bounds[0] = xmin
    bounds[1] = ymin
    bounds[2] = zmin
    bounds[3] = xmax
    bounds[4] = ymax
    bounds[5] = zmax
    return

# ==============================================================================
# =  Computation of Bounding Boxes (in toroidal configuration)
# ==============================================================================
cdef inline void comp_bbox_poly_tor(const int nvert,
                                    const double* vertr,
                                    const double* vertz,
                                    double[6] bounds) nogil:
    """
    Computes bounding box of a toroidally continous structure defined by
    the vertices vert.
    Params
    =====
        nvert : integer
           Number of vertices in the poygon
        vert : double array
           Coordinates of the polygon defining the structure in the poloidal
           plane such that vert[0:3, ii] = (x_i, y_i) the coordinates of the
           i-th vertex
        bounds : (6) double array <INOUT>
           coordinates of the lowerleftback point and of the upperrightfront
           point of the bounding box of the structure toroidally continous on
           the tore.
    """
    cdef int ii
    cdef double rmax=vertr[0], zmin=vertz[0], zmax=vertz[0]
    cdef double tmp_val
    for ii in range(1, nvert):
        tmp_val = vertr[ii]
        if tmp_val > rmax:
            rmax = tmp_val
        tmp_val = vertz[ii]
        if tmp_val > zmax:
            zmax = tmp_val
        elif tmp_val < zmin:
            zmin = tmp_val
    bounds[0] = -rmax
    bounds[1] = -rmax
    bounds[2] = zmin
    bounds[3] = rmax
    bounds[4] = rmax
    bounds[5] = zmax
    return

cdef inline void comp_bbox_poly_tor_lim(const int nvert,
                                        const double* vertr,
                                        const double* vertz,
                                        double[6] bounds,
                                        const double lmin,
                                        const double lmax) nogil:
    """
    Computes bounding box of a toroidally limited structure defined by
    the vertices vert, and limited to the angles (lmin, lmax)
    Params
    =====
    nvert : integer
       Number of vertices in the poygon
    vert : double array
       Coordinates of the polygon defining the structure in the poloidal plane
       such that vert[0:3, ii] = (x_i, y_i) the coordinates of the i-th vertex
    bounds : (6) double array <INOUT>
       coordinates of the lowerleftback point and of the upperrightfront point
       of the bounding box of the structure toroidally limited on the tore.
    lmin : double
       minimum toroidal angle where the structure lays.
    lmax : double
       maximum toroidal angle where the structure lays.
    """
    cdef int ii
    cdef double toto=100000.
    cdef double xmin=toto, xmax=-toto
    cdef double ymin=toto, ymax=-toto
    cdef double zmin=toto, zmax=-toto
    cdef double cos_min, sin_min
    cdef double cos_max, sin_max
    cdef double half_pi = 0.5 * Cpi
    cdef double[3] temp
    cdef double[6] bounds_min
    # ...
    cos_min = Ccos(lmin)
    sin_min = Csin(lmin)
    cos_max = Ccos(lmax)
    sin_max = Csin(lmax)
    if (lmin >= 0.) and (lmax >= 0.):
        if lmax > half_pi and lmin < half_pi:
            comp_bbox_poly_tor(nvert, vertr, vertz, &bounds_min[0])
            if ymax < bounds_min[4]:
                ymax = bounds_min[4]
        elif lmax < half_pi and lmin > half_pi:
            comp_bbox_poly_tor(nvert, vertr, vertz, &bounds_min[0])
            if ymin > bounds_min[1]:
                ymin = bounds_min[1]
    elif (lmin <= 0 and lmax <= 0):
        if lmax < -half_pi and lmin > -half_pi:
            comp_bbox_poly_tor(nvert, vertr, vertz, &bounds_min[0])
            if ymin > bounds_min[1]:
                ymin = bounds_min[1]
        elif lmax > -half_pi and lmin < -half_pi:
            comp_bbox_poly_tor(nvert, vertr, vertz, &bounds_min[0])
            if ymax < bounds_min[4]:
                ymax = bounds_min[4]
    elif (Cabs(Cabs(lmin) - Cpi) > _VSMALL
          and Cabs(Cabs(lmax) - Cpi) > _VSMALL):
        if lmin >= 0 :
            # lmin and lmax of opposite signs, so lmax < 0. Divide and conquer:
            comp_bbox_poly_tor_lim(nvert, vertr, vertz, &bounds[0], lmin, Cpi)
            comp_bbox_poly_tor_lim(nvert, vertr, vertz, &bounds_min[0],
                                   -Cpi, lmax)
        else:
            # lmin and lmax of opposite signs, so lmax <= 0. Divide and conquer:
            comp_bbox_poly_tor_lim(nvert, vertr, vertz, &bounds[0], lmin, -0.0)
            comp_bbox_poly_tor_lim(nvert, vertr, vertz, &bounds_min[0], 0, lmax)
        # we compute the extremes of the two boxes:
        for ii in range(3):
            if bounds[ii] > bounds_min[ii]:
                bounds[ii] = bounds_min[ii]
        for ii in range(3, 6):
            if bounds[ii] < bounds_min[ii]:
                bounds[ii] = bounds_min[ii]
        return
    for ii in range(nvert):
        temp[0] = vertr[ii]
        temp[1] = vertz[ii]
        coordshift_simple1d(temp, False, 1., cos_min, sin_min)
        # initialization:
        if xmin > temp[0]:
            xmin = temp[0]
        if xmax < temp[0]:
            xmax = temp[0]
        if ymin > temp[1]:
            ymin = temp[1]
        if ymax < temp[1]:
            ymax = temp[1]
        if zmin > temp[2]:
            zmin = temp[2]
        if zmax < temp[2]:
            zmax = temp[2]
        # .....
        temp[0] = vertr[ii]
        temp[1] = vertz[ii]
        coordshift_simple1d(temp, False, 1., cos_max, sin_max)
        if xmin > temp[0]:
            xmin = temp[0]
        if xmax < temp[0]:
            xmax = temp[0]
        if ymin > temp[1]:
            ymin = temp[1]
        if ymax < temp[1]:
            ymax = temp[1]
        if zmin > temp[2]:
            zmin = temp[2]
        if zmax < temp[2]:
            zmax = temp[2]
    bounds[0] = xmin
    bounds[1] = ymin
    bounds[2] = zmin
    bounds[3] = xmax
    bounds[4] = ymax
    bounds[5] = zmax
    return

cdef inline void coordshift_simple1d(double[3] pts, bint in_is_cartesian,
                                     double CrossRef, double cos_phi,
                                     double sin_phi) nogil:
    """
    Similar to coordshift but only pas from 3D cartesian to 3D toroidal
    coordinates or vice-versa.
    """
    cdef double x, y, z
    cdef double r, p
    if in_is_cartesian:
        if CrossRef==0.:
            x = pts[0]
            y = pts[1]
            z = pts[2]
            pts[0] = Csqrt(x*x+y*y)
            pts[1] = z
            pts[2] = Catan2(y,x)
        else:
            x = pts[0]
            y = pts[1]
            z = pts[2]
            pts[0] = Csqrt(x*x+y*y)
            pts[1] = z
            pts[2] = CrossRef
    else:
        if CrossRef==0.:
            r = pts[0]
            z = pts[1]
            p = pts[2]
            pts[0] = r*Ccos(p)
            pts[1] = r*Csin(p)
            pts[2] = z
        else:
            r = pts[0]
            z = pts[1]
            pts[0] = r*cos_phi
            pts[1] = r*sin_phi
            pts[2] = z
    return


cdef inline void is_inside_vessel(double[:, ::1] pts, double[:, ::1] ves_poly,
                                  double[:,::1] ves_lims, int nlim,
                                  bint is_toroidal, bint in_is_cartesian,
                                  int[3] order,
                                  int[::1] is_inside) nogil:
    """
    check if points are inside given vessel
    pts : points to check if inside
    ves_poly : (2, nvert) array defining polygon of the vessel
    ves_lims : (2, nlim) limits of the vessel, None if contiguous
    nlim : number of limits
    is_toroidal : true if vessel is toroidal
    in_is_cartesian : true if pts coordinates are given in cartesian coordinates
    order : order where the coordinates are stored, for example :
              - (x,y,z) => order = (0, 1, 2)
              - (z,x,y) => order = (1, 2, 0)
              - (r, z, phi) => order = (0, 1, 2)
              - (z, r, phi) => order = (1, 0, 2)
    is_inside : result, array of bools of size npts
    """
    cdef int ii, jj
    cdef int fst
    cdef int npts = pts.shape[1]
    cdef int nvert = ves_poly.shape[1] - 1
    cdef double[3] shift_pts
    cdef double[2] lims
    cdef double rii, zii, pii
    cdef bint in_ves
    # --------------------------------------------------------------------------
    if is_toroidal:
        if ves_lims is None or nlim == 0:
            if in_is_cartesian:
                for ii in range(npts):
                    shift_pts[0] = pts[order[0], ii]
                    shift_pts[1] = pts[order[1], ii]
                    shift_pts[2] = pts[order[2], ii]
                    coordshift_simple1d(shift_pts, in_is_cartesian,
                                        -1, # we dont care about phi
                                        -1, -1) # no need for these args
                    is_inside[ii] = is_point_in_path(nvert,
                                                     &ves_poly[0][0],
                                                     &ves_poly[1][0],
                                                     shift_pts[0], shift_pts[1])
            else:
                is_point_in_path_vec(nvert, &ves_poly[0][0], &ves_poly[1][0],
                                     npts, &pts[order[0]][0], &pts[order[1]][0],
                                     &is_inside[0])
        else:
            # -- There are limits ----------------------------------------------
            if in_is_cartesian:
                for jj in range(nlim):
                    fst = jj * npts
                    lims[0] = Catan2(Csin(ves_lims[jj][0]),
                                     Ccos(ves_lims[jj][0]))
                    lims[1] = Catan2(Csin(ves_lims[jj][1]),
                                     Ccos(ves_lims[jj][1]))
                    if lims[0] < lims[1]:
                        for ii in range(npts):
                            shift_pts[0] = pts[order[0], ii]
                            shift_pts[1] = pts[order[1], ii]
                            shift_pts[2] = pts[order[2], ii]
                            coordshift_simple1d(shift_pts, in_is_cartesian,
                                                0, # we need to compute phi
                                                -1, -1) # no need for these args
                            in_ves = is_point_in_path(nvert,
                                                      &ves_poly[0][0],
                                                      &ves_poly[1][0],
                                                      shift_pts[0],
                                                      shift_pts[1])
                            is_inside[fst + ii] = (in_ves
                                                   & (shift_pts[2]>=lims[0])
                                                   & (shift_pts[2]<=lims[1]))
                    else:
                        for ii in range(npts):
                            shift_pts[0] = pts[order[0], ii]
                            shift_pts[1] = pts[order[1], ii]
                            shift_pts[2] = pts[order[2], ii]
                            coordshift_simple1d(shift_pts, in_is_cartesian,
                                                0, # we need to compute phi
                                                -1, -1) # no need for these args
                            in_ves = is_point_in_path(nvert,
                                                      &ves_poly[0][0],
                                                      &ves_poly[1][0],
                                                      shift_pts[0],
                                                      shift_pts[1])
                            is_inside[fst + ii] = (in_ves
                                                   & ((shift_pts[2]>=lims[0])
                                                      | (shift_pts[2]<=lims[1])))
            else:
                # -- in coordinates: polar -------------------------------------
                for jj in range(nlim):
                    fst = jj * npts
                    lims[0] = Catan2(Csin(ves_lims[jj][0]),
                                     Ccos(ves_lims[jj][0]))
                    lims[1] = Catan2(Csin(ves_lims[jj][1]),
                                     Ccos(ves_lims[jj][1]))
                    if lims[0] < lims[1]:
                        for ii in range(npts):
                            rii = pts[order[0],ii]
                            zii = pts[order[1],ii]
                            pii = Catan2(Csin(pts[order[2],ii]),
                                         Ccos(pts[order[2],ii]))
                            in_ves = is_point_in_path(nvert,
                                                      &ves_poly[0][0],
                                                      &ves_poly[1][0],
                                                      rii, zii)
                            is_inside[fst + ii] = (in_ves
                                                   & (pii>=lims[0])
                                                   & (pii<=lims[1]))
                    else:
                        for ii in range(npts):
                            rii = pts[order[0],ii]
                            zii = pts[order[1],ii]
                            pii = Catan2(Csin(pts[order[2],ii]),
                                         Ccos(pts[order[2],ii]))
                            in_ves = is_point_in_path(nvert,
                                                      &ves_poly[0][0],
                                                      &ves_poly[1][0],
                                                      rii, zii)
                            is_inside[fst + ii] = (in_ves
                                                   & ((pii>=lims[0])
                                                      | (pii<=lims[1])))
    else:
        # -- Cylindrical case --------------------------------------------------
        if in_is_cartesian:
            for jj in range(nlim):
                fst = jj * npts
                for ii in range(npts):
                    in_ves = is_point_in_path(nvert,
                                              &ves_poly[0][0],
                                              &ves_poly[1][0],
                                              pts[order[1],ii],
                                              pts[order[2],ii])
                    is_inside[fst + ii] = (in_ves
                                           & (pts[order[0],ii]>=ves_lims[0][0])
                                           & (pts[order[0],ii]<=ves_lims[0][1]))
        else:
            # polar coordinates
            for jj in range(nlim):
                fst = jj * npts
                for ii in range(npts):
                    shift_pts[0] = pts[order[0], ii]
                    shift_pts[1] = pts[order[1], ii]
                    shift_pts[2] = pts[order[2], ii]
                    coordshift_simple1d(shift_pts, in_is_cartesian,
                                        0, # we need to compute phi
                                        -1, -1) # no need for these args
                    in_ves = is_point_in_path(nvert,
                                              &ves_poly[0][0],
                                              &ves_poly[1][0],
                                              shift_pts[1],
                                              shift_pts[2])
                    is_inside[fst + ii] = (in_ves
                                           & (shift_pts[0]>=ves_lims[0][0])
                                           & (shift_pts[0]<=ves_lims[0][1]))
    return





# ==============================================================================
# =  Raytracing basic tools: intersection ray and axis aligned bounding box
# ==============================================================================
cdef inline bint inter_ray_aabb_box(const int[3] sign,
                                    const double[3] inv_direction,
                                    const double* bounds,
                                    const double[3] ds,
                                    const bint countin) nogil:
    """
    Computes intersection between a ray (LOS) and a axis aligned bounding
    box. It returns True if ray intersects box, else False.
    Params
    =====
       sign : (3) int array
          Sign of the direction of the ray.
          If sign[i] = 1, ray_vdir[i] < 0, else sign[i] = 0
       inv_direction : (3) double array
          Inverse on each axis of direction of LOS
       bounds : (6) double array
          [3d coords of lowerleftback point of bounding box,
           3d coords of upperrightfront point of bounding box]
       ds : (3) double array
          [3d coords of origin of ray]
    Returns
    =======
       True if ray intersects bounding box, else False
    """
    cdef double tmin, tmax, tymin, tymax
    cdef double tzmin, tzmax
    cdef int t0 = 1000000
    cdef bint res

    # computing intersection
    tmin = (bounds[sign[0]*3] - ds[0]) * inv_direction[0]
    tmax = (bounds[(1-sign[0])*3] - ds[0]) * inv_direction[0]
    tymin = (bounds[(sign[1])*3 + 1] - ds[1]) * inv_direction[1]
    tymax = (bounds[(1-sign[1])*3+1] - ds[1]) * inv_direction[1]
    if ( (tmin > tymax) or (tymin > tmax) ):
        return 0
    if (tymin > tmin):
        tmin = tymin
    if (tymax < tmax):
        tmax = tymax
    if not inv_direction[2] == 1./_VSMALL:
        tzmin = (bounds[(sign[2])*3+2] - ds[2]) * inv_direction[2]
        tzmax = (bounds[(1-sign[2])*3+2] - ds[2]) * inv_direction[2]
    else:
        tzmin = Cnan
        tzmax = Cnan
    if ( (tmin > tzmax) or (tzmin > tmax) ):
        return 0
    if (tzmin > tmin):
        tmin = tzmin
    if (tzmax < tmax):
        tmax = tzmax
    if countin and (tmin < 0.) and (tmax < 0.):
        return 0
    elif not countin and tmin < 0:
        return 0

    res = (tmin < t0) and (tmax > -t0)
    return  res


# ==============================================================================
# =  Raytracing basic tools: intersection ray and triangle (in 3d space)
# ==============================================================================
cdef inline bint inter_ray_triangle(const double[3] ray_orig,
                                    const double[3] ray_vdir,
                                    const double* vert0,
                                    const double* vert1,
                                    const double* vert2) nogil:
    cdef int ii
    cdef double det, invdet, u, v
    cdef double[3] edge1, edge2
    cdef double[3] pvec, tvec, qvec
    #...
    for ii in range(3):
        edge1[ii] = vert1[ii] - vert0[ii]
        edge2[ii] = vert2[ii] - vert0[ii]
    # begin calculating determinant  also used to calculate U parameter
    _bgt.compute_cross_prod(ray_vdir, edge2, pvec)
    # if determinant is near zero ray lies in plane of triangle
    det = _bgt.compute_dot_prod(edge1, pvec)
    if Cabs(det) < _VSMALL:
        return False
    invdet = 1./det
    # calculate distance from vert to ray origin
    for ii in range(3):
        tvec[ii] = ray_orig[ii] - vert0[ii]
    # calculate U parameter and test bounds
    u = _bgt.compute_dot_prod(tvec, pvec) * invdet
    if u < 0. or u > 1.:
        return False
    # prepare to test V parameter
    _bgt.compute_cross_prod(tvec, edge1, qvec)
    # calculate V parameter and test bounds
    v = _bgt.compute_dot_prod(ray_vdir, qvec) * invdet
    if v < 0. or u + v > 1.:
        return False
    return True

# ==============================================================================
# =  Raytracing on a Torus
# ==============================================================================
cdef inline void raytracing_inout_struct_tor(const int num_los,
                                             const double[:,::1] ray_vdir,
                                             const double[:,::1] ray_orig,
                                             double[::1] coeff_inter_out,
                                             double[::1] coeff_inter_in,
                                             double[::1] vperp_out,
                                             const long[::1] lstruct_nlim,
                                             int[::1] ind_inter_out,
                                             const bint forbid0,
                                             const bint forbidbis_org,
                                             const double rmin,
                                             const double rmin2,
                                             const double crit2_base,
                                             const int nstruct_lim,
                                             const double* lbounds,
                                             const double* langles,
                                             const int* lis_limited,
                                             const long* lnvert,
                                             const long* lsz_lim,
                                             const double* lstruct_polyx,
                                             const double* lstruct_polyy,
                                             const double* lstruct_normx,
                                             const double* lstruct_normy,
                                             const double eps_uz,
                                             const double eps_vz,
                                             const double eps_a,
                                             const double eps_b,
                                             const double eps_plane,
                                             const int num_threads,
                                             const bint is_out_struct) nogil:
    """
    Computes the entry and exit point of all provided LOS/rays for a set of
    structures that can be of type "OUT" (is_out_struct=True) or "IN"
    (is_out_struct=False) in a TORE. An "OUT" structure cannot be penetrated
    whereas an "IN" structure can. The latter is typically a vessel and are
    toroidally continous. If a structure is limited we can determine the number
    of limits and the limits itself. For optimization reasons we will also pass
    the bounding box limits. And the information of the last intersected point,
    if any.
    This functions is parallelized.

    Params
    ======
    num_los : int
       Total number of lines of sight (LOS) (aka. rays)
    ray_vdir : (3, num_los) double array
       LOS normalized direction vector
    ray_orig : (3, num_los) double array
       LOS origin points coordinates
    coeff_inter_out : (num_los) double array <INOUT>
       Coefficient of exit (kout) of the last point of intersection for each LOS
       with the global geometry (with ALL structures)
    coeff_inter_in : (num_los) double array <INOUT>
       Coefficient of entry (kin) of the last point of intersection for each LOS
       with the global geometry (with ALL structures). If intersection at origin
       k = 0, if no intersection k = NAN
    vperp_out : (3*num_los) double array <INOUT>
       Coordinates of the normal vector of impact of the LOS (0 if none). It is
       stored in the following way [v_{0,x}, v_{0,y}, v_{0,z}, ..., v_{n-1,z}]
    lstruct_nlim : array of ints
       List of number of limits for all structures
    ind_inter_out : (3 * num_los)  <INOUT>
       Index of structure impacted by LOS such that:
                ind_inter_out[ind_los*3:ind_los*3+3]=(i,j,k)
       where k is the index of edge impacted on the j-th sub structure of the
       structure number i. If the LOS impacted the vessel i=j=0
    forbid0 : bool
       Should we forbid values behind visible radius ? (see Rmin). If false,
       will test "hidden" part always, else, it will depend on the LOS and
       on forbidbis.
    forbidbis: bint
       Should we forbid values behind visible radius for each LOS ?
    rmin : double
       Minimal radius of vessel to take into consideration
    rmin2 : double
       Squared valued of the minimal radius
    crit2_base : double
       Critical value to evaluate for each LOS if horizontal or not
    nstruct_lim : int
       Number of OUT structures (not counting the limited versions).
       If not is_out_struct then length of vpoly.
    lbounds : (6 * nstruct) double array
       Coordinates of lower and upper edges of the bounding box for each
       structures (nstruct = sum_i(nstruct_lim * lsz_lim[i])
       If not is_out_struct then NULL
    langles : (2 * nstruct) double array
       Minimum and maximum angles where the structure lives. If the structure
       number 'i' is toroidally continous then langles[i:i+2] = [0, 0].
    lis_limited : (nstruct) int array
       List of bool to know if the structures (or the vessel) is limited or not.
    lnvert : (nstruct_lim) int array
       List of vertices of each polygon for each structure
       If not is_out_struct then NULL
    lsz_lim : (nstruct) int array
       List of the total number of structures before the ith structure. First
       element is always 0, else lsz_lim[i] = sum_j(lstruct_nlim[j], j=0..i-1)
       If not is_out_struct then NULL
    lstruct_polyx : (ntotnvert)
       List of "x" coordinates of the polygon's vertices of all structures on
       the poloidal plane
    lstruct_polyy : (ntotnvert)
       List of "y" coordinates of the polygon's vertices of all structures on
       the poloidal plane
    lstruct_normx : (2, num_vertex-1) double array
       List of "x" coordinates of the normal vectors going "inwards" of the
        edges of the Polygon defined by lstruct_poly
    lstruct_normy : (2, num_vertex-1) double array
       List of "y" coordinates of the normal vectors going "inwards" of the
        edges of the Polygon defined by lstruct_poly
    eps<val> : double
       Small value, acceptance of error
    num_threads : int
       The num_threads argument indicates how many threads the team should
       consist of. If not given, OpenMP will decide how many threads to use.
       Typically this is the number of cores available on the machine.
    is_out_struct : bint
       Bool to determine if the structure is "OUT" or "IN". An "OUT" structure
       cannot be penetrated whereas an "IN" structure can. The latter is
       typically a vessel and are toroidally continous.
    """
    cdef double upscaDp=0., upar2=0., dpar2=0., crit2=0., idpar2=0.
    cdef double dist = 0., s1x = 0., s1y = 0., s2x = 0., s2y = 0.
    cdef double lim_min=0., lim_max=0., invuz=0.
    cdef int totnvert=0
    cdef int nvert
    cdef int ind_struct, ind_bounds
    cdef int ind_los, ii, jj, kk
    cdef bint lim_is_none
    cdef bint found_new_kout
    cdef bint inter_bbox
    cdef bint forbidbis = forbidbis_org
    cdef double* last_pout = NULL
    cdef double* kpout_loc = NULL
    cdef double* kpin_loc = NULL
    cdef double* invr_ray = NULL
    cdef double* loc_org = NULL
    cdef double* loc_dir = NULL
    cdef double* lim_ves = NULL
    cdef double* loc_vp = NULL
    cdef int* sign_ray = NULL
    cdef int* ind_loc = NULL

    # == Defining parallel part ================================================
    with nogil, parallel(num_threads=num_threads):
        # We use local arrays for each thread so
        loc_org   = <double *> malloc(sizeof(double) * 3)
        loc_dir   = <double *> malloc(sizeof(double) * 3)
        loc_vp    = <double *> malloc(sizeof(double) * 3)
        kpin_loc  = <double *> malloc(sizeof(double) * 1)
        kpout_loc = <double *> malloc(sizeof(double) * 1)
        ind_loc   = <int *> malloc(sizeof(int) * 1)
        if is_out_struct:
            # if the structure is "out" (solid) we need more arrays
            last_pout = <double *> malloc(sizeof(double) * 3)
            invr_ray  = <double *> malloc(sizeof(double) * 3)
            lim_ves   = <double *> malloc(sizeof(double) * 2)
            sign_ray  = <int *> malloc(sizeof(int) * 3)

        # == The parallelization over the LOS ==================================
        for ind_los in prange(num_los, schedule='dynamic'):
            ind_struct = 0
            loc_org[0] = ray_orig[0, ind_los]
            loc_org[1] = ray_orig[1, ind_los]
            loc_org[2] = ray_orig[2, ind_los]
            loc_dir[0] = ray_vdir[0, ind_los]
            loc_dir[1] = ray_vdir[1, ind_los]
            loc_dir[2] = ray_vdir[2, ind_los]
            loc_vp[0] = 0.
            loc_vp[1] = 0.
            loc_vp[2] = 0.
            if is_out_struct:
                # if structure is of "Out" type, then we compute the last
                # point where it went out of a structure.
                ind_loc[0] = ind_inter_out[2+3*ind_los]
                kpin_loc[0] = coeff_inter_out[ind_los]
                last_pout[0] = kpin_loc[0] * loc_dir[0] + loc_org[0]
                last_pout[1] = kpin_loc[0] * loc_dir[1] + loc_org[1]
                last_pout[2] = kpin_loc[0] * loc_dir[2] + loc_org[2]
                compute_inv_and_sign(loc_dir, sign_ray, invr_ray)
            else:
                kpout_loc[0] = 0
                kpin_loc[0] = 0
                ind_loc[0] = 0

            # -- Computing values that depend on the LOS/ray -------------------
            upscaDp = loc_dir[0]*loc_org[0] + loc_dir[1]*loc_org[1]
            upar2   = loc_dir[0]*loc_dir[0] + loc_dir[1]*loc_dir[1]
            dpar2   = loc_org[0]*loc_org[0] + loc_org[1]*loc_org[1]
            idpar2 = 1./dpar2
            invuz = 1./loc_dir[2]
            crit2 = upar2*crit2_base

            # -- Prepare in case forbid is True --------------------------------
            if forbid0 and not dpar2>0:
                forbidbis = 0
            if forbidbis:
                # Compute coordinates of the 2 points where the tangents touch
                # the inner circle
                dist = Csqrt(dpar2-rmin2)
                s1x = (rmin2 * loc_org[0] + rmin * loc_org[1] * dist) * idpar2
                s1y = (rmin2 * loc_org[1] - rmin * loc_org[0] * dist) * idpar2
                s2x = (rmin2 * loc_org[0] - rmin * loc_org[1] * dist) * idpar2
                s2y = (rmin2 * loc_org[1] + rmin * loc_org[0] * dist) * idpar2

            # == Case "OUT" structure ==========================================
            if is_out_struct:
                # We work on each structure
                for ii in range(nstruct_lim):
                    # -- Getting structure's data ------------------------------
                    if ii == 0:
                        nvert = lnvert[0]
                        totnvert = 0
                    else:
                        totnvert = lnvert[ii-1]
                        nvert = lnvert[ii] - totnvert
                    ind_struct = lsz_lim[ii]
                    # -- Working on the structure limited ----------------------
                    for jj in range(lstruct_nlim[ii]):
                        lim_min = langles[(ind_struct+jj)*2]
                        lim_max = langles[(ind_struct+jj)*2 + 1]
                        lim_is_none = lis_limited[ind_struct+jj] == 1
                        # We test if it is really necessary to compute the inter
                        # ie. we check if the ray intersects the bounding box

                        inter_bbox = inter_ray_aabb_box(sign_ray, invr_ray,
                                                        &lbounds[(ind_struct
                                                                  + jj)*6],
                                                        loc_org,
                                                        True)
                        if not inter_bbox:
                            continue
                        # We check that the bounding box is not "behind"
                        # the last POut encountered
                        inter_bbox = inter_ray_aabb_box(sign_ray, invr_ray,
                                                        &lbounds[(ind_struct
                                                                  + jj)*6],
                                                        last_pout, False)
                        if inter_bbox:
                            continue
                         # Else, we compute the new values
                        found_new_kout \
                            = comp_inter_los_vpoly(loc_org,
                                                   loc_dir,
                                                   &lstruct_polyx[totnvert],
                                                   &lstruct_polyy[totnvert],
                                                   &lstruct_normx[totnvert-ii],
                                                   &lstruct_normy[totnvert-ii],
                                                   nvert-1,
                                                   lim_is_none,
                                                   lim_min, lim_max,
                                                   forbidbis,
                                                   upscaDp, upar2,
                                                   dpar2, invuz,
                                                   s1x, s1y,
                                                   s2x, s2y,
                                                   crit2, eps_uz,
                                                   eps_vz, eps_a,
                                                   eps_b, eps_plane,
                                                   False,
                                                   kpin_loc,
                                                   kpout_loc,
                                                   ind_loc,
                                                   loc_vp)
                        if found_new_kout :
                            coeff_inter_out[ind_los] = kpin_loc[0]
                            vperp_out[0+3*ind_los] = loc_vp[0]
                            vperp_out[1+3*ind_los] = loc_vp[1]
                            vperp_out[2+3*ind_los] = loc_vp[2]
                            ind_inter_out[2+3*ind_los] = ind_loc[0]
                            ind_inter_out[0+3*ind_los] = 1+ii
                            ind_inter_out[1+3*ind_los] = jj
                            last_pout[0] = (coeff_inter_out[ind_los] *
                                            loc_dir[0]) + loc_org[0]
                            last_pout[1] = (coeff_inter_out[ind_los] *
                                            loc_dir[1]) + loc_org[1]
                            last_pout[2] = (coeff_inter_out[ind_los] *
                                            loc_dir[2]) + loc_org[2]
            else:
                # == Case "IN" structure =======================================
                # Nothing to do but compute intersection between vessel and LOS
                found_new_kout = comp_inter_los_vpoly(loc_org, loc_dir,
                                                      lstruct_polyx,
                                                      lstruct_polyy,
                                                      lstruct_normx,
                                                      lstruct_normy,
                                                      nstruct_lim,
                                                      lis_limited[0],
                                                      langles[0], langles[1],
                                                      forbidbis,
                                                      upscaDp, upar2,
                                                      dpar2, invuz,
                                                      s1x, s1y, s2x, s2y,
                                                      crit2, eps_uz, eps_vz,
                                                      eps_a,eps_b, eps_plane,
                                                      True,
                                                      kpin_loc, kpout_loc,
                                                      ind_loc, loc_vp)
                if found_new_kout:
                    coeff_inter_in[ind_los]  = kpin_loc[0]
                    coeff_inter_out[ind_los] = kpout_loc[0]
                    ind_inter_out[2+3*ind_los] = ind_loc[0]
                    ind_inter_out[0+3*ind_los] = 0
                    ind_inter_out[1+3*ind_los] = 0
                    vperp_out[0+3*ind_los] = loc_vp[0]
                    vperp_out[1+3*ind_los] = loc_vp[1]
                    vperp_out[2+3*ind_los] = loc_vp[2]

                else:
                    coeff_inter_in[ind_los]  = Cnan
                    coeff_inter_out[ind_los] = Cnan
                    ind_inter_out[2+3*ind_los] = 0
                    ind_inter_out[0+3*ind_los] = 0
                    ind_inter_out[1+3*ind_los] = 0
                    vperp_out[0+3*ind_los] = 0.
                    vperp_out[1+3*ind_los] = 0.
                    vperp_out[2+3*ind_los] = 0.
            # end case IN/OUT
        free(loc_org)
        free(loc_dir)
        free(loc_vp)
        free(kpin_loc)
        free(kpout_loc)
        free(ind_loc)
        if is_out_struct:
            free(last_pout)
            free(lim_ves)
            free(invr_ray)
            free(sign_ray)
    return


# ------------------------------------------------------------------
cdef inline bint comp_inter_los_vpoly(const double[3] ray_orig,
                                      const double[3] ray_vdir,
                                      const double* lpolyx,
                                      const double* lpolyy,
                                      const double* normx,
                                      const double* normy,
                                      const int nvert,
                                      const bint lim_is_none,
                                      const double lim_min,
                                      const double lim_max,
                                      const bint forbidbis,
                                      const double upscaDp, const double upar2,
                                      const double dpar2, const double invuz,
                                      const double s1x,   const double s1y,
                                      const double s2x, const double s2y,
                                      const double crit2, const double eps_uz,
                                      const double eps_vz, const double eps_a,
                                      const double eps_b, const double eps_pln,
                                      const bint is_in_struct,
                                      double[1] kpin_loc, double[1] kpout_loc,
                                      int[1] ind_loc, double[3] vperpin) nogil:
    """
    Computes the entry and exit point of ONE provided LOS/rays for a single
    structure that can be of type "OUT" (is_out_struct=True) or "IN"
    (is_out_struct=False). An "OUT" structure cannot be penetrated whereas an
    "IN" structure can. The latter is typically a vessel and are toroidally
    continous. If a structure is limited we can determine the number of limits
    and the limits itself. For optimization reasons we will also pass the
    bounding box limits. And the information of the last intersected point, if
    any.

    Params
    ======
    ray_vdir : (3) double array
       LOS normalized direction vector
    ray_orig : (3) double array
       LOS origin points coordinates
    lpolyx : (ntotnvert)
       List of "x" coordinates of the polygon's vertices of the structures on
       the poloidal plane
    lpolyy : (ntotnvert)
       List of "y" coordinates of the polygon's vertices of the structures on
       the poloidal plane
    normx : (2, num_vertex-1) double array
       List of "x" coordinates of the normal vectors going "inwards" of the
        edges of the Polygon defined by lpolyx/y
    normy : (2, num_vertex-1) double array
       List of "y" coordinates of the normal vectors going "inwards" of the
        edges of the Polygon defined by lpolyxy
    nvert : int
       Number of vertices on the polygon
    lim_is_none : bint
       Bool to know if the structures (or the vessel) is limited or not.
    lim_min : double
       Minimum angle where the structure lives. If the structure
       is toroidally continous then lim_min = 0
    lim_max : double
       Maximum angle where the structure lives. If the structure
       is toroidally continous then lim_min = 0
    forbidbis: bint
       Should we forbid values behind visible radius for each LOS ?
    upscaDp: double
       Scalar product between LOS' origin and direction
    upar2 : double
       Norm direction LOS
    dpar2 : double
       Norm origin LOS
    invuz : double
       Inverse of 3rd component of direction. ie. if direction is (ux, uy, uz)
       then invuz = 1/uz
    s1x, s1y, s2x, s2y : double
       Compute coordinates of the 2 points where the tangents touch the inner
       circle of the Tore, only needed if forbidbis = 0
    crit2 : double
       Critical value to evaluate for each LOS if horizontal or not
    eps<val> : double
       Small value, acceptance of error
    is_in_struct : bint
       Bool to determine if the structure is "OUT" or "IN". An "OUT" structure
       cannot be penetrated whereas an "IN" structure can. The latter is
       typically a vessel and are toroidally continous.
    kpout_loc : double array <INOUT>
       Coefficient of exit (kout) of the last point of intersection for the LOS
       with the structure or vessel
    kpin_loc : double array <INOUT>
       Coefficient of exit (kin) of the last point of intersection for the LOS
       with the structure or vessel
    vperpin : (3) double array <INOUT>
       Coordinates of the normal vector of impact of the LOS (0 if none)
    Return
    ======
    bool : If true, there was in impact
           If false, no intersection between LOS and structure
    """
    cdef int jj
    cdef int done=0
    cdef int indin=0
    cdef int indout=0
    cdef bint inter_bbox
    cdef double kout, kin
    cdef double res_kin = kpin_loc[0]
    cdef double res_kout = kpout_loc[0]
    cdef double sca=0., sca0=0., sca1=0., sca2=0.
    cdef double q, coeff, delta, sqd, k, sol0, sol1, phi=0.
    cdef double v0, v1, val_a, val_b, ephi_in0, ephi_in1
    cdef double sout1, sout0
    cdef double sin1, sin0
    cdef double invupar2
    cdef double cosl0, cosl1, sinl0, sinl1
    cdef double[3] opp_dir

    # -- Computing some seful values -------------------------------------------
    cosl0 = Ccos(lim_min)
    cosl1 = Ccos(lim_max)
    sinl0 = Csin(lim_min)
    sinl1 = Csin(lim_max)
    invupar2 = 1./upar2
    # == Compute all solutions =================================================
    # Set tolerance value for ray_vdir[2,ii]
    # eps_uz is the tolerated DZ across 20m (max Tokamak size)
    kout, kin, done = 1.e12, 1.e12, 0
    if ray_vdir[2] * ray_vdir[2] < crit2:
        # -- Case with horizontal semi-line ------------------------------------
        for jj in range(nvert):
            # Solutions exist only in the case with non-horizontal
            # segment (i.e.: cone, not plane)
            if (lpolyy[jj+1] - lpolyy[jj])**2 > eps_vz * eps_vz:
                q = (ray_orig[2] - lpolyy[jj]) / (lpolyy[jj+1] - lpolyy[jj])
                # The intersection must stand on the segment
                if q>=0 and q<1:
                    coeff = q * q * (lpolyx[jj+1]-lpolyx[jj])**2 + \
                        2. * q * lpolyx[jj] * (lpolyx[jj+1] - lpolyx[jj]) + \
                        lpolyx[jj] * lpolyx[jj]
                    delta = upscaDp * upscaDp - upar2 * (dpar2 - coeff)
                    if delta>0.:
                        sqd = Csqrt(delta)
                        # The intersection must be on the semi-line (i.e.: k>=0)
                        # First solution
                        if -upscaDp - sqd >= 0:
                            k = (-upscaDp - sqd) * invupar2
                            sol0 = ray_orig[0] + k * ray_vdir[0]
                            sol1 = ray_orig[1] + k * ray_vdir[1]
                            if forbidbis:
                                sca0 = (sol0-s1x)*ray_orig[0] + \
                                       (sol1-s1y)*ray_orig[1]
                                sca1 = (sol0-s1x)*s1x + (sol1-s1y)*s1y
                                sca2 = (sol0-s2x)*s2x + (sol1-s2y)*s2y
                            if not forbidbis or (forbidbis and not
                                                 (sca0<0 and sca1<0 and
                                                  sca2<0)):
                                # Get the normalized perpendicular vector
                                # at intersection
                                phi = Catan2(sol1, sol0)
                                # Check sol inside the Lim
                                if lim_is_none or (not lim_is_none and
                                                   ((lim_min<lim_max and
                                                     lim_min<=phi and
                                                     phi<=lim_max)
                                                    or (lim_min>lim_max and
                                                        (phi>=lim_min or
                                                         phi<=lim_max)))):
                                    # Get the scalar product to determine
                                    # entry or exit point
                                    sca = Ccos(phi)*normx[jj]*ray_vdir[0] + \
                                          Csin(phi)*normx[jj]*ray_vdir[1] + \
                                          normy[jj]*ray_vdir[2]
                                    if sca<=0 and k<kout:
                                        kout = k
                                        done = 1
                                        indout = jj
                                    elif sca>=0 and k<min(kin,kout):
                                        kin = k
                                        indin = jj

                        # Second solution
                        if -upscaDp + sqd >=0:
                            k = (-upscaDp + sqd)*invupar2
                            sol0 = ray_orig[0] + k * ray_vdir[0]
                            sol1 = ray_orig[1] + k * ray_vdir[1]
                            if forbidbis:
                                sca0 = (sol0-s1x) * ray_orig[0] + \
                                       (sol1-s1y) * ray_orig[1]
                                sca1 = (sol0-s1x) * s1x + (sol1-s1y) * s1y
                                sca2 = (sol0-s2x) * s2x + (sol1-s2y) * s2y
                            if not forbidbis or (forbidbis and not
                                                 (sca0<0 and sca1<0 and
                                                  sca2<0)):
                                # Get the normalized perpendicular vector
                                # at intersection
                                phi = Catan2(sol1,sol0)
                                if lim_is_none or (not lim_is_none and
                                                   ((lim_min<lim_max and
                                                     lim_min<=phi and
                                                     phi<=lim_max) or
                                                    (lim_min>lim_max and
                                                     (phi>=lim_min or
                                                      phi<=lim_max))
                                                   )):
                                    # Get the scalar product to determine
                                    # entry or exit point
                                    sca = Ccos(phi)*normx[jj]*ray_vdir[0] + \
                                          Csin(phi)*normx[jj]*ray_vdir[1] + \
                                          normy[jj]*ray_vdir[2]
                                    if sca<=0 and k<kout:
                                        kout = k
                                        done = 1
                                        indout = jj
                                    elif sca>=0 and k<min(kin,kout):
                                        kin = k
                                        indin = jj
    else:
        # == More general non-horizontal semi-line case ========================
        for jj in range(nvert):
            v0 = lpolyx[jj+1]-lpolyx[jj]
            v1 = lpolyy[jj+1]-lpolyy[jj]
            val_a = v0 * v0 - upar2 * v1 * invuz * v1 * invuz
            val_b = lpolyx[jj] * v0 + v1 * (ray_orig[2] - lpolyy[jj]) * upar2 *\
                    invuz * invuz - upscaDp * v1 * invuz
            coeff = - upar2 * (ray_orig[2] - lpolyy[jj])**2 * invuz * invuz +\
                    2. * upscaDp * (ray_orig[2]-lpolyy[jj]) * invuz -\
                    dpar2 + lpolyx[jj] * lpolyx[jj]
            if ((val_a * val_a < eps_a * eps_a) and
                (val_b * val_b > eps_b * eps_b)):
                q = -coeff / (2. * val_b)
                if q >= 0. and q < 1.:
                    k = (q * v1 - (ray_orig[2] - lpolyy[jj])) * invuz
                    if k >= 0:
                        sol0 = ray_orig[0] + k * ray_vdir[0]
                        sol1 = ray_orig[1] + k * ray_vdir[1]
                        if forbidbis:
                            sca0 = (sol0-s1x)*ray_orig[0] + \
                                   (sol1-s1y)*ray_orig[1]
                            sca1 = (sol0-s1x)*s1x + (sol1-s1y)*s1y
                            sca2 = (sol0-s2x)*s2x + (sol1-s2y)*s2y
                            if sca0<0 and sca1<0 and sca2<0:
                                continue
                        # Get the normalized perpendicular vect at intersection
                        phi = Catan2(sol1,sol0)
                        if lim_is_none or (not lim_is_none and
                                           ((lim_min < lim_max and
                                             lim_min <= phi and
                                             phi <= lim_max) or
                                            (lim_min > lim_max and
                                             (phi >= lim_min or
                                              phi <= lim_max)))):
                            # Get the scal prod to determine entry or exit point
                            sca = Ccos(phi) * normx[jj] * ray_vdir[0] + \
                                  Csin(phi) * normx[jj] * ray_vdir[1] + \
                                  normy[jj] * ray_vdir[2]
                            if sca<=0 and k<kout:
                                kout = k
                                done = 1
                                indout = jj
                            elif sca>=0 and k<min(kin,kout):
                                kin = k
                                indin = jj
            elif ((val_a * val_a >= eps_a * eps_a) and
                  (val_b * val_b > val_a * coeff)):
                sqd = Csqrt(val_b * val_b - val_a * coeff)
                # First solution
                q = (-val_b + sqd) / val_a
                if q >= 0. and q < 1.:
                    k = (q * v1 - (ray_orig[2] - lpolyy[jj])) * invuz
                    if k >= 0.:
                        sol0 = ray_orig[0] + k * ray_vdir[0]
                        sol1 = ray_orig[1] + k * ray_vdir[1]
                        if forbidbis:
                            sca0 = (sol0-s1x) * ray_orig[0] + \
                                   (sol1-s1y) * ray_orig[1]
                            sca1 = (sol0-s1x) * s1x + (sol1-s1y) * s1y
                            sca2 = (sol0-s2x) * s2x + (sol1-s2y) * s2y
                        if not forbidbis or (forbidbis and
                                             not (sca0<0 and sca1<0 and
                                                  sca2<0)):
                            # Get the normalized perpendicular vector at inter
                            phi = Catan2(sol1, sol0)
                            if lim_is_none or (not lim_is_none and
                                               ((lim_min < lim_max and
                                                 lim_min <= phi and
                                                 phi <= lim_max) or
                                                (lim_min > lim_max and
                                                 (phi >= lim_min or
                                                  phi <= lim_max)))):
                                # Get the scal prod to determine in or out point
                                sca = Ccos(phi) * normx[jj] * ray_vdir[0] + \
                                      Csin(phi) * normx[jj] * ray_vdir[1] + \
                                      normy[jj] * ray_vdir[2]
                                if sca<=0 and k<kout:
                                    kout = k
                                    done = 1
                                    indout = jj
                                elif sca>=0 and k<min(kin,kout):
                                    kin = k
                                    indin = jj

                # == Second solution ===========================================
                q = (-val_b - sqd) / val_a
                if q >= 0. and q < 1.:
                    k = (q * v1 - (ray_orig[2] - lpolyy[jj])) * invuz
                    if k>=0.:
                        sol0 = ray_orig[0] + k * ray_vdir[0]
                        sol1 = ray_orig[1] + k * ray_vdir[1]
                        if forbidbis:
                            sca0 = (sol0-s1x) * ray_orig[0] + \
                                   (sol1-s1y) * ray_orig[1]
                            sca1 = (sol0-s1x) * s1x + (sol1-s1y) * s1y
                            sca2 = (sol0-s2x) * s2x + (sol1-s2y) * s2y
                        if not forbidbis or (forbidbis and
                                             not (sca0<0 and sca1<0 and
                                                  sca2<0)):
                            # Get the normalized perpendicular vector at inter
                            phi = Catan2(sol1,sol0)
                            if lim_is_none or (not lim_is_none and
                                               ((lim_min < lim_max and
                                                 lim_min <= phi and
                                                 phi <= lim_max) or
                                                (lim_min>lim_max and
                                                 (phi>=lim_min or
                                                  phi<=lim_max)))):
                                # Get the scal prod to determine if in or out
                                sca = Ccos(phi) * normx[jj] * ray_vdir[0] + \
                                      Csin(phi) * normx[jj] * ray_vdir[1] + \
                                      normy[jj] * ray_vdir[2]
                                if sca<=0 and k<kout:
                                    kout = k
                                    done = 1
                                    indout = jj
                                elif sca>=0 and k<min(kin,kout):
                                    kin = k
                                    indin = jj

    if not lim_is_none:
        ephi_in0 = -sinl0
        ephi_in1 =  cosl0
        if Cabs(ray_vdir[0] * ephi_in0 + ray_vdir[1] * ephi_in1) > eps_pln:
            k = -(ray_orig[0] * ephi_in0 + ray_orig[1] * ephi_in1) \
                /(ray_vdir[0] * ephi_in0 + ray_vdir[1] * ephi_in1)
            if k >= 0:
                # Check if in ves_poly
                sol0 = (ray_orig[0] + k * ray_vdir[0]) * cosl0 + \
                       (ray_orig[1] + k * ray_vdir[1]) * sinl0
                sol1 =  ray_orig[2] + k * ray_vdir[2]
                inter_bbox = is_point_in_path(nvert, lpolyx, lpolyy, sol0, sol1)
                if inter_bbox:
                    # Check PIn (POut not possible for limited torus)
                    sca = ray_vdir[0] * ephi_in0 + ray_vdir[1] * ephi_in1
                    if sca<=0 and k<kout:
                        kout = k
                        done = 1
                        indout = -1
                    elif sca>=0 and k<min(kin,kout):
                        kin = k
                        indin = -1

        ephi_in0 =  sinl1
        ephi_in1 = -cosl1
        if Cabs(ray_vdir[0] * ephi_in0 + ray_vdir[1] * ephi_in1) > eps_pln:
            k = -(ray_orig[0] * ephi_in0 + ray_orig[1] * ephi_in1)\
                /(ray_vdir[0] * ephi_in0 + ray_vdir[1] * ephi_in1)
            if k >= 0:
                sol0 = (ray_orig[0] + k * ray_vdir[0]) * cosl1 +\
                       (ray_orig[1] + k * ray_vdir[1]) * sinl1
                sol1 =  ray_orig[2] + k * ray_vdir[2]
                # Check if in ves_poly
                inter_bbox = is_point_in_path(nvert, lpolyx, lpolyy, sol0, sol1)
                if inter_bbox:
                    # Check PIn (POut not possible for limited torus)
                    sca = ray_vdir[0]*ephi_in0 + ray_vdir[1]*ephi_in1
                    if sca<=0 and k<kout:
                        kout = k
                        done = 1
                        indout = -2
                    elif sca>=0 and k<min(kin,kout):
                        kin = k
                        indin = -2
    # == Analyzing if there was impact =========================================
    if done==1:
        if is_in_struct :
            kpout_loc[0] = kout
            if indout==-1:
                vperpin[0] = -sinl0
                vperpin[1] = cosl0
                vperpin[2] = 0.
            elif indout==-2:
                vperpin[0] = sinl1
                vperpin[1] = -cosl1
                vperpin[2] = 0.
            else:
                sout0 = ray_orig[0] + kout * ray_vdir[0]
                sout1 = ray_orig[1] + kout * ray_vdir[1]
                phi = Catan2(sout1, sout0)
                vperpin[0] = Ccos(phi) * normx[indout]
                vperpin[1] = Csin(phi) * normx[indout]
                vperpin[2] = normy[indout]
            ind_loc[0] = indout
            if kin<kout:
                kpin_loc[0] = kin
        elif kin < kout and kin < res_kin:
            kpin_loc[0] = kin
            if indin==-1:
                vperpin[0] = sinl0
                vperpin[1] = -cosl0
                vperpin[2] = 0.
            elif indin==-2:
                vperpin[0] = -sinl1
                vperpin[1] = cosl1
                vperpin[2] = 0.
            else:
                sin0 = ray_orig[0] + kin * ray_vdir[0]
                sin1 = ray_orig[1] + kin * ray_vdir[1]
                phi = Catan2(sin1,sin0)
                vperpin[0] = -Ccos(phi) * normx[indin]
                vperpin[1] = -Csin(phi) * normx[indin]
                vperpin[2] = -normy[indin]
            ind_loc[0] = indin
    return (res_kin != kpin_loc[0]) or (res_kout != kpout_loc[0]
                                        and is_in_struct)

# ==============================================================================
# =  Raytracing on a Cylinder (Linear case)
# ==============================================================================
cdef inline void raytracing_inout_struct_lin(const int Nl,
                                             const double[:,::1] Ds,
                                             const double [:,::1] us,
                                             const int Ns,
                                             const double* polyx_tab,
                                             const double* polyy_tab,
                                             const double* normx_tab,
                                             const double* normy_tab,
                                             const double L0,
                                             const double L1,
                                             double[::1] kin_tab,
                                             double[::1] kout_tab,
                                             double[::1] vperpout_tab,
                                             int[::1] indout_tab,
                                             const double EpsPlane,
                                             const int ind_struct,
                                             const int ind_lim_struct) nogil:

    cdef bint is_in_path
    cdef int ii=0, jj=0
    cdef double kin, kout, scauVin, q, X, sca, k
    cdef int indin=0, indout=0, Done=0

    if ind_struct == 0 and ind_lim_struct == 0 :
        # If it is the first struct,
        # we have to initialize values even if no impact
        kin_tab[ii]  = Cnan
        kout_tab[ii] = Cnan

    for ii in range(0,Nl):
        kout, kin, Done = 1.e12, 1e12, 0
        # For cylinder
        for jj in range(0,Ns):
            scauVin = us[1,ii] * normx_tab[jj] + us[2,ii] * normy_tab[jj]
            # Only if plane not parallel to line
            if Cabs(scauVin)>EpsPlane:
                k = -( (Ds[1,ii] - polyx_tab[jj]) * normx_tab[jj] +
                       (Ds[2,ii] - polyy_tab[jj]) * normy_tab[jj]) \
                       / scauVin
                # Only if on good side of semi-line
                if k>=0.:
                    V1 = polyx_tab[jj+1]-polyx_tab[jj]
                    V2 = polyy_tab[jj+1]-polyy_tab[jj]
                    if (V1**2 + V2**2 > _VSMALL):
                        q = (  (Ds[1,ii] + k * us[1,ii] - polyx_tab[jj]) * V1
                             + (Ds[2,ii] + k * us[2,ii] - polyy_tab[jj]) * V2) \
                             / (V1*V1 + V2*V2)
                        # Only of on the fraction of plane
                        if q>=0. and q<1.:
                            X = Ds[0,ii] + k*us[0,ii]
                            # Only if within limits
                            if X>=L0 and X<=L1:
                                sca = us[1,ii] * normx_tab[jj] \
                                      + us[2,ii] * normy_tab[jj]
                                # Only if new
                                if sca<=0 and k<kout:
                                    kout = k
                                    indout = jj
                                    Done = 1
                                elif sca>=0 and k<min(kin,kout):
                                    kin = k
                                    indin = jj
                    else:
                        with gil:
                            from warnings import warn
                            warn("The polygon has double identical points",
                                 Warning)
        # For two faces
        # Only if plane not parallel to line
        if Cabs(us[0,ii])>EpsPlane:
            # First face
            k = -(Ds[0,ii]-L0)/us[0,ii]
            # Only if on good side of semi-line
            if k>=0.:
                # Only if inside VPoly
                is_in_path = is_point_in_path(Ns, polyx_tab, polyy_tab,
                                              Ds[1,ii]+k*us[1,ii],
                                              Ds[2,ii]+k*us[2,ii])
                if is_in_path:
                    if us[0,ii]<=0 and k<kout:
                        kout = k
                        indout = -1
                        Done = 1
                    elif us[0,ii]>=0 and k<min(kin,kout):
                        kin = k
                        indin = -1
            # Second face
            k = -(Ds[0,ii]-L1)/us[0,ii]
            # Only if on good side of semi-line
            if k>=0.:
                # Only if inside VPoly
                is_in_path = is_point_in_path(Ns, polyx_tab, polyy_tab,
                                              Ds[1,ii]+k*us[1,ii],
                                              Ds[2,ii]+k*us[2,ii])
                if is_in_path:
                    if us[0,ii]>=0 and k<kout:
                        kout = k
                        indout = -2
                        Done = 1
                    elif us[0,ii]<=0 and k<min(kin,kout):
                        kin = k
                        indin = -2
        # == Analyzing if there was impact ====================================
        if Done==1:
            if (ind_struct == 0 and ind_lim_struct == 0):
                kout_tab[ii] = kout
                if kin < kout:
                    kin_tab[ii] = kin
                # To be finished
                if indout==-1:
                    vperpout_tab[0 + 3 * ii] = 1.
                    vperpout_tab[1 + 3 * ii] = 0.
                    vperpout_tab[2 + 3 * ii] = 0.
                elif indout==-2:
                    vperpout_tab[0 + 3 * ii] = -1.
                    vperpout_tab[1 + 3 * ii] = 0.
                    vperpout_tab[2 + 3 * ii] = 0.
                else:
                    vperpout_tab[0 + 3 * ii] = 0.
                    vperpout_tab[1 + 3 * ii] = normx_tab[indout]
                    vperpout_tab[2 + 3 * ii] = normy_tab[indout]
                indout_tab[0 + 3 * ii] = 0
                indout_tab[1 + 3 * ii] = 0
                indout_tab[2 + 3 * ii] = indout
            elif kin<kout_tab[ii] and kin < kout:
                kout_tab[ii] = kin
                indout_tab[0 + 3 * ii] = ind_struct
                indout_tab[1 + 3 * ii] = ind_lim_struct
                indout_tab[2 + 3 * ii] = indin
                if indout==-1:
                    vperpout_tab[0 + 3 * ii] = 1.
                    vperpout_tab[1 + 3 * ii] = 0.
                    vperpout_tab[2 + 3 * ii] = 0.
                elif indout==-2:
                    vperpout_tab[0 + 3 * ii] = -1.
                    vperpout_tab[1 + 3 * ii] = 0.
                    vperpout_tab[2 + 3 * ii] = 0.
                else:
                    vperpout_tab[0 + 3 * ii] = 0.
                    vperpout_tab[1 + 3 * ii] = normx_tab[indout]
                    vperpout_tab[2 + 3 * ii] = normy_tab[indout]
    return

cdef inline void compute_inout_tot(const int num_los,
                                   const int npts_poly,
                                   const double[:, ::1] ray_orig,
                                   const double[:, ::1] ray_vdir,
                                   const double[:, ::1] ves_poly,
                                   const double[:, ::1] ves_norm,
                                   const long[::1] lstruct_nlim_org,
                                   const double[::1] ves_lims,
                                   const double[::1] lstruct_polyx,
                                   const double[::1] lstruct_polyy,
                                   list lstruct_lims,
                                   const double[::1] lstruct_normx,
                                   const double[::1] lstruct_normy,
                                   const long[::1] lnvert,
                                   const int nstruct_tot,
                                   const int nstruct_lim,
                                   const int sz_ves_lims,
                                   const double min_poly_r,
                                   const double rmin_org,
                                   const double eps_uz, const double eps_a,
                                   const double eps_vz, const double eps_b,
                                   const double eps_plane, str ves_type,
                                   const bint forbid, const int num_threads,
                                   double[::1] coeff_inter_out,
                                   double[::1] coeff_inter_in,
                                   double[::1] vperp_out,
                                   int[::1] ind_inter_out) :
    cdef int ii, jj, kk
    cdef int ind_struct = 0
    cdef int len_lim
    cdef int ind_min
    cdef int nvert
    cdef double crit2_base = eps_uz * eps_uz /400.
    cdef double lim_min = 0.
    cdef double lim_max = 0.
    cdef double rmin = rmin_org
    cdef double rmin2 = 0.
    cdef bint forbidbis, forbid0
    cdef double *lbounds = <double *>malloc(nstruct_tot * 6 * sizeof(double))
    cdef double *langles = <double *>malloc(nstruct_tot * 2 * sizeof(double))
    cdef int *llimits = NULL
    cdef long *lsz_lim = NULL
    cdef int[1] llim_ves
    cdef double[2] lbounds_ves
    cdef double[2] lim_ves
    cdef long[::1] lstruct_nlim
    # ==========================================================================
    if ves_type == 'tor':
        # .. if there are, we get the limits for the vessel ....................
        if ves_lims is None or sz_ves_lims == 0:
            are_limited = False
            lbounds_ves[0] = 0
            lbounds_ves[1] = 0
            llim_ves[0] = 1
        else:
            are_limited = True
            lbounds_ves[0] = Catan2(Csin(ves_lims[0]), Ccos(ves_lims[0]))
            lbounds_ves[1] = Catan2(Csin(ves_lims[1]), Ccos(ves_lims[1]))
            llim_ves[0] = 0
        # -- Toroidal case -----------------------------------------------------
        # rmin is necessary to avoid looking on the other side of the tokamak
        if rmin < 0.:
            rmin = 0.95*min(min_poly_r,
                            _bgt.comp_min_hypot(ray_orig[0, ...],
                                                ray_orig[1, ...],
                                                num_los))
        rmin2 = rmin*rmin
        # Variable to avoid looking "behind" blind spot of tore
        if forbid:
            forbid0, forbidbis = 1, 1
        else:
            forbid0, forbidbis = 0, 0
        # -- Computing intersection between LOS and Vessel ---------------------
        raytracing_inout_struct_tor(num_los, ray_vdir, ray_orig,
                                    coeff_inter_out, coeff_inter_in,
                                    vperp_out, None, ind_inter_out,
                                    forbid0, forbidbis,
                                    rmin, rmin2, crit2_base,
                                    npts_poly,  NULL, lbounds_ves,
                                    llim_ves, NULL, NULL,
                                    &ves_poly[0][0],
                                    &ves_poly[1][0],
                                    &ves_norm[0][0],
                                    &ves_norm[1][0],
                                    eps_uz, eps_vz, eps_a, eps_b, eps_plane,
                                    num_threads, False) # structure is in
        # -- Treating the structures (if any) ----------------------------------
        if nstruct_tot > 0:
            ind_struct = 0
            llimits = <int *>malloc(nstruct_tot * sizeof(int))
            lsz_lim = <long *>malloc(nstruct_lim * sizeof(long))
            lstruct_nlim = lstruct_nlim_org.copy()
            for ii in range(nstruct_lim):
                # For fast accessing
                len_lim = lstruct_nlim[ii]
                # We get the limits if any
                if len_lim == 0:
                    lslim = [None]
                    lstruct_nlim[ii] = lstruct_nlim[ii] + 1
                elif len_lim == 1:
                    lslim = [[lstruct_lims[ii][0, 0], lstruct_lims[ii][0, 1]]]
                else:
                    lslim = lstruct_lims[ii]
                # We get the number of vertices and limits of the struct's poly
                if ii == 0:
                    lsz_lim[0] = 0
                    nvert = lnvert[0]
                    ind_min = 0
                else:
                    nvert = lnvert[ii] - lnvert[ii - 1]
                    lsz_lim[ii] = lstruct_nlim[ii-1] + lsz_lim[ii-1]
                    ind_min = lnvert[ii-1]
                # and loop over the limits (one continous structure)
                for jj in range(max(len_lim,1)):
                    # We compute the structure's bounding box:
                    if lslim[jj] is not None:
                        lim_ves[0] = lslim[jj][0]
                        lim_ves[1] = lslim[jj][1]
                        llimits[ind_struct] = 0 # False : struct is limited
                        lim_min = Catan2(Csin(lim_ves[0]), Ccos(lim_ves[0]))
                        lim_max = Catan2(Csin(lim_ves[1]), Ccos(lim_ves[1]))
                        comp_bbox_poly_tor_lim(nvert,
                                               &lstruct_polyx[ind_min],
                                               &lstruct_polyy[ind_min],
                                               &lbounds[ind_struct*6],
                                               lim_min, lim_max)
                    else:
                        llimits[ind_struct] = 1 # True : is continous
                        comp_bbox_poly_tor(nvert,
                                           &lstruct_polyx[ind_min],
                                           &lstruct_polyy[ind_min],
                                           &lbounds[ind_struct*6])
                        lim_min = 0.
                        lim_max = 0.
                    langles[ind_struct*2] = lim_min
                    langles[ind_struct*2 + 1] = lim_max
                    ind_struct = 1 + ind_struct
            # end loops over structures
            # -- Computing intersection between structures and LOS -------------
            raytracing_inout_struct_tor(num_los, ray_vdir, ray_orig,
                                        coeff_inter_out, coeff_inter_in,
                                        vperp_out, lstruct_nlim,
                                        ind_inter_out,
                                        forbid0, forbidbis,
                                        rmin, rmin2, crit2_base,
                                        nstruct_lim,
                                        lbounds, langles, llimits,
                                        &lnvert[0], lsz_lim,
                                        &lstruct_polyx[0],
                                        &lstruct_polyy[0],
                                        &lstruct_normx[0],
                                        &lstruct_normy[0],
                                        eps_uz, eps_vz, eps_a,
                                        eps_b, eps_plane,
                                        num_threads,
                                        True) # the structure is "OUT"
            free(lsz_lim)
            free(llimits)
    else:
        # -- Cylindrical case --------------------------------------------------
        # .. if there are, we get the limits for the vessel ....................
        if ves_lims is None  or sz_ves_lims == 0:
            are_limited = False
            lbounds_ves[0] = 0
            lbounds_ves[1] = 0
        else:
            are_limited = True
            lbounds_ves[0] = ves_lims[0]
            lbounds_ves[1] = ves_lims[1]

        raytracing_inout_struct_lin(num_los, ray_orig, ray_vdir, npts_poly,
                                    &ves_poly[0][0], &ves_poly[1][0],
                                    &ves_norm[0][0], &ves_norm[1][0],
                                    lbounds_ves[0], lbounds_ves[1],
                                    coeff_inter_in, coeff_inter_out,
                                    vperp_out, ind_inter_out, eps_plane,
                                    0, 0) # The vessel is strcuture 0,0
        # -- Treating the structures (if any) ----------------------------------
        if nstruct_tot > 0:
            ind_struct = 0
            lstruct_nlim = lstruct_nlim_org.copy()
            for ii in range(nstruct_lim):
                # -- Analyzing the limits --------------------------------------
                len_lim = lstruct_nlim[ii]
                # We get the limits if any
                if len_lim == 0:
                    lslim = [None]
                    lstruct_nlim[ii] = lstruct_nlim[ii] + 1
                elif len_lim == 1:
                    lslim = [[lstruct_lims[ii][0, 0], lstruct_lims[ii][0, 1]]]
                else:
                    lslim = lstruct_lims[ii]
                if ii == 0:
                    nvert = lnvert[0]
                    ind_min = 0
                else:
                    nvert = lnvert[ii] - lnvert[ii - 1]
                    ind_min = lnvert[ii-1]
                # and loop over the limits (one continous structure)
                for jj in range(max(len_lim,1)):
                    if lslim[jj] is not None:
                        lbounds_ves[0] = lslim[jj][0]
                        lbounds_ves[1] = lslim[jj][1]
                    raytracing_inout_struct_lin(num_los, ray_orig, ray_vdir,
                                                nvert-1,
                                                &lstruct_polyx[ind_min],
                                                &lstruct_polyy[ind_min],
                                                &lstruct_normx[ind_min-ii],
                                                &lstruct_normy[ind_min-ii],
                                                lbounds_ves[0],
                                                lbounds_ves[1],
                                                coeff_inter_in,
                                                coeff_inter_out,
                                                vperp_out, ind_inter_out,
                                                eps_plane, ii+1, jj)
    free(lbounds)
    free(langles)
    return


# ==============================================================================
# =  Raytracing on a Torus only KMin and KMax
# ==============================================================================
cdef inline void raytracing_minmax_struct_tor(const int num_los,
                                              const double[:,::1] ray_vdir,
                                              const double[:,::1] ray_orig,
                                              double* coeff_inter_out,
                                              double* coeff_inter_in,
                                              const bint forbid0,
                                              const bint forbidbis_org,
                                              const double rmin, double rmin2,
                                              const double crit2_base,
                                              const int npts_poly,
                                              const double* langles,
                                              const bint is_limited,
                                              const double* surf_polyx,
                                              const double* surf_polyy,
                                              const double* surf_normx,
                                              const double* surf_normy,
                                              const double eps_uz,
                                              const double eps_vz,
                                              const double eps_a,
                                              const double eps_b,
                                              const double eps_plane,
                                              const int num_threads) nogil:
    """
    Computes the entry and exit point of all provided LOS/rays for a set of
    "IN" structures in a TORE. A "in" structure is typically a vessel, or
    flux surface and are (noramally) toroidally continous but you can specify
    if it is otherwise with lis_limited and langles.
    This functions is parallelized.

    Params
    ======
    num_los : int
       Total number of lines of sight (LOS) (aka. rays)
    ray_vdir : (3, num_los) double array
       LOS normalized direction vector
    ray_orig : (3, num_los) double array
       LOS origin points coordinates
    coeff_inter_out : (num_los*num_surf) double array <INOUT>
       Coefficient of exit (kout) of the last point of intersection for each LOS
       with the global geometry (with ALL structures)
    coeff_inter_in : (num_los*num_surf) double array <INOUT>
       Coefficient of entry (kin) of the last point of intersection for each LOS
       with the global geometry (with ALL structures). If intersection at origin
       k = 0, if no intersection k = NAN
    forbid0 : bool
       Should we forbid values behind visible radius ? (see Rmin). If false,
       will test "hidden" part always, else, it will depend on the LOS and
       on forbidbis.
    forbidbis: bint
       Should we forbid values behind visible radius for each LOS ?
    rmin : double
       Minimal radius of vessel to take into consideration
    rmin2 : double
       Squared valued of the minimal radius
    crit2_base : double
       Critical value to evaluate for each LOS if horizontal or not
    npts_poly : int
       Number of OUT structures (not counting the limited versions).
       If not is_out_struct then length of vpoly.
    langles : (2 * nstruct) double array
       Minimum and maximum angles where the structure lives. If the structure
       number 'i' is toroidally continous then langles[i:i+2] = [0, 0].
    is_limited : bint
       bool to know if the flux surface is limited or not
    surf_polyx : (ntotnvert)
       List of "x" coordinates of the polygon's vertices on
       the poloidal plane
    surf_polyy : (ntotnvert)
       List of "y" coordinates of the polygon's vertices on
       the poloidal plane
    surf_normx : (2, num_vertex-1) double array
       List of "x" coordinates of the normal vectors going "inwards" of the
        edges of the Polygon defined by surf_poly
    surf_normy : (2, num_vertex-1) double array
       List of "y" coordinates of the normal vectors going "inwards" of the
        edges of the Polygon defined by surf_poly
    eps<val> : double
       Small value, acceptance of error
    num_threads : int
       The num_threads argument indicates how many threads the team should
       consist of. If not given, OpenMP will decide how many threads to use.
       Typically this is the number of cores available on the machine.
    """
    cdef double upscaDp=0., upar2=0., dpar2=0., crit2=0., idpar2=0.
    cdef double dist = 0., s1x = 0., s1y = 0., s2x = 0., s2y = 0.
    cdef double lim_min=0., lim_max=0., invuz=0.
    cdef int totnvert=0
    cdef int nvert
    cdef int ind_struct, ind_bounds
    cdef int ind_los, ii, jj, kk
    cdef bint lim_is_none
    cdef bint found_new_kout
    cdef bint forbidbis = forbidbis_org
    cdef double[3] dummy
    cdef int[1] silly
    cdef double* kpout_loc = NULL
    cdef double* kpin_loc = NULL
    cdef double* loc_org = NULL
    cdef double* loc_dir = NULL

    # == Defining parallel part ================================================
    with nogil, parallel(num_threads=num_threads):
        # We use local arrays for each thread so
        loc_org   = <double *> malloc(sizeof(double) * 3)
        loc_dir   = <double *> malloc(sizeof(double) * 3)
        kpin_loc  = <double *> malloc(sizeof(double) * 1)
        kpout_loc = <double *> malloc(sizeof(double) * 1)
        # == The parallelization over the LOS ==================================
        for ind_los in prange(num_los, schedule='dynamic'):
            ind_struct = 0
            loc_org[0] = ray_orig[0, ind_los]
            loc_org[1] = ray_orig[1, ind_los]
            loc_org[2] = ray_orig[2, ind_los]
            loc_dir[0] = ray_vdir[0, ind_los]
            loc_dir[1] = ray_vdir[1, ind_los]
            loc_dir[2] = ray_vdir[2, ind_los]
            kpout_loc[0] = 0
            kpin_loc[0] = 0
            # -- Computing values that depend on the LOS/ray -------------------
            upscaDp = loc_dir[0]*loc_org[0] + loc_dir[1]*loc_org[1]
            upar2   = loc_dir[0]*loc_dir[0] + loc_dir[1]*loc_dir[1]
            dpar2   = loc_org[0]*loc_org[0] + loc_org[1]*loc_org[1]
            idpar2 = 1./dpar2
            invuz = 1./loc_dir[2]
            crit2 = upar2*crit2_base

            # -- Prepare in case forbid is True --------------------------------
            if forbid0 and not dpar2>0:
                forbidbis = 0
            if forbidbis:
                # Compute coordinates of the 2 points where the tangents touch
                # the inner circle
                dist = Csqrt(dpar2-rmin2)
                s1x = (rmin2 * loc_org[0] + rmin * loc_org[1] * dist) * idpar2
                s1y = (rmin2 * loc_org[1] - rmin * loc_org[0] * dist) * idpar2
                s2x = (rmin2 * loc_org[0] - rmin * loc_org[1] * dist) * idpar2
                s2y = (rmin2 * loc_org[1] + rmin * loc_org[0] * dist) * idpar2

            # == Case "IN" structure =======================================
            # Nothing to do but compute intersection between vessel and LOS
            found_new_kout = comp_inter_los_vpoly(loc_org, loc_dir,
                                                  surf_polyx,
                                                  surf_polyy,
                                                  surf_normx,
                                                  surf_normy,
                                                  npts_poly-1,
                                                  not is_limited,
                                                  langles[0], langles[1],
                                                  forbidbis,
                                                  upscaDp, upar2,
                                                  dpar2, invuz,
                                                  s1x, s1y, s2x, s2y,
                                                  crit2, eps_uz, eps_vz,
                                                  eps_a,eps_b, eps_plane,
                                                  True,
                                                  kpin_loc, kpout_loc,
                                                  silly, dummy)
            if found_new_kout:
                coeff_inter_in[ind_los]  = kpin_loc[0]
                coeff_inter_out[ind_los] = kpout_loc[0]
            else:
                coeff_inter_in[ind_los]  = Cnan
                coeff_inter_out[ind_los] = Cnan
        free(loc_org)
        free(loc_dir)
        free(kpin_loc)
        free(kpout_loc)
    return


# ==============================================================================
# =  Raytracing on a Cylinder only KMin and KMax
# ==============================================================================
cdef inline void raytracing_minmax_struct_lin(const int Nl,
                                              const double[:,::1] Ds,
                                              const double [:,::1] us,
                                              const int Ns,
                                              const double* polyx_tab,
                                              const double* polyy_tab,
                                              const double* normx_tab,
                                              const double* normy_tab,
                                              const double L0,
                                              const double L1,
                                              double* kin_tab,
                                              double* kout_tab,
                                              const double EpsPlane) nogil:
    cdef bint is_in_path
    cdef int ii=0, jj=0
    cdef double kin, kout, scauVin, q, X, sca, k
    cdef int indin=0, indout=0, Done=0

    kin_tab[ii]  = Cnan
    kout_tab[ii] = Cnan

    for ii in range(0,Nl):
        kout = 1.e12
        kin  = 1.e12
        Done = 0
        # For cylinder
        for jj in range(0,Ns):
            scauVin = us[1,ii] * normx_tab[jj] + us[2,ii] * normy_tab[jj]
            # Only if plane not parallel to line
            if Cabs(scauVin)>EpsPlane:
                k = -( (Ds[1,ii] - polyx_tab[jj]) * normx_tab[jj] +
                       (Ds[2,ii] - polyy_tab[jj]) * normy_tab[jj]) \
                       / scauVin
                # Only if on good side of semi-line
                if k>=0.:
                    V1 = polyx_tab[jj+1]-polyx_tab[jj]
                    V2 = polyy_tab[jj+1]-polyy_tab[jj]
                    q = (  (Ds[1,ii] + k * us[1,ii] - polyx_tab[jj]) * V1
                         + (Ds[2,ii] + k * us[2,ii] - polyy_tab[jj]) * V2) \
                         / (V1*V1 + V2*V2)
                    # Only of on the fraction of plane
                    if q>=0. and q<1.:
                        X = Ds[0,ii] + k*us[0,ii]

                        # Only if within limits
                        if X>=L0 and X<=L1:
                            sca = us[1,ii] * normx_tab[jj] \
                                  + us[2,ii] * normy_tab[jj]
                            # Only if new
                            if sca<=0 and k<kout:
                                kout = k
                                indout = jj
                                Done = 1
                            elif sca>=0 and k<min(kin,kout):
                                kin = k
                                indin = jj

        # For two faces
        # Only if plane not parallel to line
        if Cabs(us[0,ii])>EpsPlane:
            # First face
            k = -(Ds[0,ii]-L0)/us[0,ii]
            # Only if on good side of semi-line
            if k>=0.:
                # Only if inside VPoly
                is_in_path = is_point_in_path(Ns, polyx_tab, polyy_tab,
                                              Ds[1,ii]+k*us[1,ii],
                                              Ds[2,ii]+k*us[2,ii])
                if is_in_path:
                    if us[0,ii]<=0 and k<kout:
                        kout = k
                        indout = -1
                        Done = 1
                    elif us[0,ii]>=0 and k<min(kin,kout):
                        kin = k
                        indin = -1
            # Second face
            k = -(Ds[0,ii]-L1)/us[0,ii]
            # Only if on good side of semi-line
            if k>=0.:
                # Only if inside VPoly
                is_in_path = is_point_in_path(Ns, polyx_tab, polyy_tab,
                                              Ds[1,ii]+k*us[1,ii],
                                              Ds[2,ii]+k*us[2,ii])
                if is_in_path:
                    if us[0,ii]>=0 and k<kout:
                        kout = k
                        indout = -2
                        Done = 1
                    elif us[0,ii]<=0 and k<min(kin,kout):
                        kin = k
                        indin = -2
        # == Analyzing if there was impact ====================================
        if Done==1:
            kout_tab[ii] = kout
            if kin<kin_tab[ii]:
                kin_tab[ii] = kin
    return

# ==============================================================================
# = Checking if points are visible
# ==============================================================================
cdef inline void is_visible_pt_vec(double pt0, double pt1, double pt2,
                                   double[:,::1] pts, int npts,
                                   double[:, ::1] ves_poly,
                                   double[:, ::1] ves_norm,
                                   double[::1] ind,
                                   double[::1] k,
                                   double[::1] ves_lims,
                                   long[::1] lstruct_nlim,
                                   double[::1] lstruct_polyx,
                                   double[::1] lstruct_polyy,
                                   list lstruct_lims,
                                   double[::1] lstruct_normx,
                                   double[::1] lstruct_normy,
                                   long[::1] lnvert,
                                   int nstruct_tot,
                                   int nstruct_lim,
                                   double rmin,
                                   double eps_uz, double eps_a,
                                   double eps_vz, double eps_b,
                                   double eps_plane, str ves_type,
                                   bint forbid,
                                   bint test, int num_threads):
    cdef array vperp_out = clone(array('d'), npts * 3, True)
    cdef array coeff_inter_in  = clone(array('d'), npts, True)
    cdef array coeff_inter_out = clone(array('d'), npts, True)
    cdef array ind_inter_out = clone(array('i'), npts * 3, True)
    cdef int sz_ves_lims
    cdef double min_poly_r
    cdef double* dist_arr = NULL
    cdef double[:,::1] ray_orig = view.array(shape=(3,npts),
                                             itemsize=sizeof(double),
                                             format="d")
    cdef double[:,::1] ray_vdir = view.array(shape=(3,npts),
                                             itemsize=sizeof(double),
                                             format="d")
    cdef int npts_poly = ves_norm.shape[1]
    # --------------------------------------------------------------------------
    # Initialization : creation of the rays between points pts and P
    _bgt.tile_3_to_2d(pt0, pt1, pt2, npts, ray_orig)
    if k == None:
        dist_arr = <double*> malloc(npts*sizeof(double))
        _bgt.compute_dist_pt_vec(pt0, pt1, pt2, npts, pts, &dist_arr[0])
        _bgt.compute_diff_div(pts, ray_orig, dist_arr, npts, ray_vdir)
    else:
        _bgt.compute_diff_div(pts, ray_orig, &k[0], npts, ray_vdir)
    # --------------------------------------------------------------------------
    sz_ves_lims = np.size(ves_lims)
    min_poly_r = _bgt.comp_min(ves_poly[0, ...], npts_poly-1)
    compute_inout_tot(npts, npts_poly,
                      ray_orig, ray_vdir,
                      ves_poly, ves_norm,
                      lstruct_nlim, ves_lims,
                      lstruct_polyx, lstruct_polyy,
                      lstruct_lims, lstruct_normx,
                      lstruct_normy, lnvert,
                      nstruct_tot, nstruct_lim,
                      sz_ves_lims, min_poly_r, rmin,
                      eps_uz, eps_a, eps_vz, eps_b,
                      eps_plane, ves_type,
                      forbid, num_threads,
                      coeff_inter_out, coeff_inter_in, vperp_out,
                      ind_inter_out)
    # --------------------------------------------------------------------------
    # Get ind
    if k == None:
        is_vis_mask(ind, dist_arr, coeff_inter_out, npts)
        free(dist_arr)
    else:
        is_vis_mask(ind, &k[0], coeff_inter_out, npts)
    return

cdef inline void is_vis_mask(double[::1] ind, double* k,
                             double[::1] coeff_inter_out,
                             int npts) nogil:
    cdef int ii
    for ii in range(npts):
        ind[ii] = 1
        if k[ii] > coeff_inter_out[ii]:
            ind[ii] = 0
    return

cdef inline void are_visible_vec_vec(double[:, ::1] pts1, int npts1,
                                     double[:,::1] pts2, int npts2,
                                     double[:, ::1] ves_poly,
                                     double[:, ::1] ves_norm,
                                     double[:, ::1] ind,
                                     double[::1] k,
                                     double[::1] ves_lims,
                                     long[::1] lstruct_nlim,
                                     double[::1] lstruct_polyx,
                                     double[::1] lstruct_polyy,
                                     list lstruct_lims,
                                     double[::1] lstruct_normx,
                                     double[::1] lstruct_normy,
                                     long[::1] lnvert,
                                     int nstruct_tot,
                                     int nstruct_lim,
                                     double rmin,
                                     double eps_uz, double eps_a,
                                     double eps_vz, double eps_b,
                                     double eps_plane, str ves_type,
                                     bint forbid,
                                     bint test, int num_threads):
    cdef np.ndarray[double, ndim=2, mode='c'] ray_orig_arr
    cdef np.ndarray[double, ndim=2, mode='c'] ray_vdir_arr
    cdef np.ndarray[double, ndim=2, mode='c'] dist_arr
    cdef double[:,::1] ray_orig
    cdef double[:,::1] ray_vdir
    cdef int ii
    # We compute for each point in the polygon
    for ii in range(npts1):
        is_visible_pt_vec(pts1[0,ii], pts1[1,ii], pts1[2,ii],
                          pts2, npts2,
                          ves_poly, ves_norm,
                          ind[ii,...], k, ves_lims,
                          lstruct_nlim,
                          lstruct_polyx, lstruct_polyy,
                          lstruct_lims,
                          lstruct_normx, lstruct_normy,
                          lnvert, nstruct_tot, nstruct_lim,
                          rmin, eps_uz, eps_a, eps_vz, eps_b,
                          eps_plane, ves_type,
                          forbid, test, num_threads)
    return




# Built-in
import os
import warnings

# Common
import numpy as np
import scipy.optimize as scpopt
import scipy.constants as scpct
import scipy.sparse as sparse
from scipy.interpolate import BSpline
import matplotlib.pyplot as plt


#                   --------------
#       TO BE MOVED TO tofu.data WHEN FINISHED !!!!
#                   --------------


_NPEAKMAX = 12

###########################################################
###########################################################
#
#           Preliminary
#       utility tools for 1d spectral fitting
#
###########################################################
###########################################################


def remove_bck(x, y):
    # opt = np.polyfit(x, y, deg=0)
    opt = [np.nanmin(y)]
    return y-opt[0], opt[0]


def get_peaks(x, y, nmax=None):

    if nmax is None:
        nmax = _NPEAKMAX

    # Prepare
    ybis = np.copy(y)
    A = np.empty((nmax,), dtype=y.dtype)
    x0 = np.empty((nmax,), dtype=x.dtype)
    sigma = np.empty((nmax,), dtype=y.dtype)
    def gauss(xx, A, x0, sigma): return A*np.exp(-(xx-x0)**2/sigma**2)
    def gauss_jac(xx, A, x0, sigma):
        jac = np.empty((xx.size, 3), dtype=float)
        jac[:, 0] = np.exp(-(xx-x0)**2/sigma**2)
        jac[:, 1] = A*2*(xx-x0)/sigma**2 * np.exp(-(xx-x0)**2/sigma**2)
        jac[:, 2] = A*2*(xx-x0)**2/sigma**3 * np.exp(-(xx-x0)**2/sigma**2)
        return jac

    dx = np.nanmin(np.diff(x))

    # Loop
    nn = 0
    while nn < nmax:
        ind = np.nanargmax(ybis)
        x00 = x[ind]
        if np.any(np.diff(ybis[ind:], n=2) >= 0.):
            wp = min(x.size-1,
                     ind + np.nonzero(np.diff(ybis[ind:],n=2)>=0.)[0][0] + 1)
        else:
            wp = ybis.size-1
        if np.any(np.diff(ybis[:ind+1], n=2) >= 0.):
            wn = max(0, np.nonzero(np.diff(ybis[:ind+1],n=2)>=0.)[0][-1] - 1)
        else:
            wn = 0
        width = x[wp]-x[wn]
        assert width>0.
        indl = np.arange(wn, wp+1)
        sig = np.ones((indl.size,))
        if (np.abs(np.mean(np.diff(ybis[ind:wp+1])))
            > np.abs(np.mean(np.diff(ybis[wn:ind+1])))):
            sig[indl < ind] = 1.5
            sig[indl > ind] = 0.5
        else:
            sig[indl < ind] = 0.5
            sig[indl > ind] = 1.5
        p0 = (ybis[ind], x00, width)#,0.)
        bounds = (np.r_[0., x[wn], dx/2.],
                  np.r_[5.*ybis[ind], x[wp], 5.*width])
        try:
            (Ai, x0i, sigi) = scpopt.curve_fit(gauss, x[indl], ybis[indl],
                                               p0=p0, bounds=bounds, jac=gauss_jac,
                                               sigma=sig, x_scale='jac')[0]
        except Exception as err:
            print(str(err))
            import ipdb
            ipdb.set_trace()
            pass

        ybis = ybis - gauss(x, Ai, x0i, sigi)
        A[nn] = Ai
        x0[nn] = x0i
        sigma[nn] = sigi


        nn += 1
    return A, x0, sigma

def get_p0bounds_all(x, y, nmax=None, lamb0=None):

    yflat, bck = remove_bck(x, y)
    amp, x0, sigma = get_peaks(x, yflat, nmax=nmax)
    lamb0 = x0
    nmax = lamb0.size

    p0 = amp.tolist() + [0 for ii in range(nmax)] + sigma.tolist() + [bck]

    lx = [np.nanmin(x), np.nanmax(x)]
    Dx = np.diff(lx)
    dx = np.nanmin(np.diff(x))

    bamp = (np.zeros(nmax,), np.full((nmax,),3.*np.nanmax(y)))
    bdlamb = (np.full((nmax,), -Dx/2.), np.full((nmax,), Dx/2.))
    bsigma = (np.full((nmax,), dx/2.), np.full((nmax,), Dx/2.))
    bbck0 = (0., np.nanmax(y))

    bounds = (np.r_[bamp[0], bdlamb[0], bsigma[0], bbck0[0]],
              np.r_[bamp[1], bdlamb[1], bsigma[1], bbck0[1]])
    if not np.all(bounds[0]<bounds[1]):
        msg = "Lower bounds must be < upper bounds !\n"
        msg += "    lower :  %s\n"+str(bounds[0])
        msg += "    upper :  %s\n"+str(bounds[1])
        raise Exception(msg)
    return p0, bounds, lamb0

def get_p0bounds_lambfix(x, y, nmax=None, lamb0=None):

    nmax = lamb0.size
    # get typical x units
    Dx = x.nanmax()-x.nanmin()
    dx = np.nanmin(np.diff(x))

    # Get background and background-subtracted y
    yflat, bck = remove_bck(x, y)

    # get initial guesses
    amp = [yflat[np.nanargmin(np.abs(x-lamb))] for lamb in lamb0]
    sigma = [Dx/nmax for ii in range(nmax)]
    p0 = A + sigma + [bck]

    # Get bounding boxes
    bamp = (np.zeros(nmax,), np.full((nmax,),3.*np.nanmax(y)))
    bsigma = (np.full((nmax,), dx/2.), np.full((nmax,), Dx/2.))
    bbck0 = (0., np.nanmax(y))

    bounds = (np.r_[bamp[0], bsigma[0], bbck0[0]],
              np.r_[bamp[1], bsigma[1], bbck0[1]])
    if not np.all(bounds[0]<bounds[1]):
        msg = "Lower bounds must be < upper bounds !\n"
        msg += "    lower :  %s\n"+str(bounds[0])
        msg += "    upper :  %s\n"+str(bounds[1])
        raise Exception(msg)
    return p0, bounds, lamb0


def get_func1d_all(n=5, lamb0=None):
    if lamb0 is None:
        lamb0 = np.zeros((n,), dtype=float)
    assert lamb0.size == n

    def func_vect(x, amp, dlamp, sigma, bck0, lamb0=lamb0, n=n):
        y = np.full((n+1, x.size), np.nan)
        y[:-1, :] = amp[:, None]*np.exp(-(x[None, :]-(lamb0+dlamb)[:, None])**2
                                        /sigma[:, None]**2)
        y[-1, :] = bck0
        return y

    def func_sca(x, *args, lamb0=lamb0, n=n):
        amp = np.r_[args[0:n]][:, None]
        dlamb = np.r_[args[n:2*n]][:, None]
        sigma = np.r_[args[2*n:3*n]][:, None]
        bck0 = np.r_[args[3*n]]
        gaus = amp * np.exp(-(x[None, :]-(lamb0[:, None] + dlamb))**2/sigma**2)
        back = bck0
        return np.sum(gaus, axis=0) + back

    def func_sca_jac(x, *args, lamb0=lamb0, n=n):
        amp = np.r_[args[0:n]][None, :]
        dlamb = np.r_[args[n:2*n]][None, :]
        sigma = np.r_[args[2*n:3*n]][None, :]
        bck0 = np.r_[args[3*n]]
        lamb0 = lamb0[None, :]
        x = x[:, None]
        jac = np.full((x.size, 3*n+1,), np.nan)
        jac[:, :n] = np.exp(-(x - (lamb0+dlamb))**2/sigma**2)
        jac[:, n:2*n] = amp*2*((x - (lamb0+dlamb))/(sigma**2)
                               * np.exp(-(x - (lamb0+dlamb))**2/sigma**2))
        jac[:, 2*n:3*n] = amp*2*((x - (lamb0+dlamb))**2/sigma**3
                                 * np.exp(-(x - (lamb0+dlamb))**2/sigma**2))
        jac[:, -1] = 1.
        return jac

    return func_vect, func_sca, func_sca_jac


def get_func1d_lamb0fix(n=5, lamb0=None):
    if lamb0 is None:
        lamb0 = np.zeros((n,), dtype=float)
    assert lamb0.size == n

    def func_vect(x, amp, sigma, bck0, lamb0=lamb0, n=n):
        y = np.full((n+1, x.size), np.nan)
        for ii in range(n):
            y[ii, :] = amp[ii]*np.exp(-(x-lamb0[ii])**2/sigma[ii]**2)
        y[-1, :] = bck0
        return y

    def func_sca(x, *args, lamb0=lamb0, n=n):
        amp = np.r_[args[0:n]][:, None]
        sigma = np.r_[args[2*n:3*n]][:, None]
        bck0 = np.r_[args[3*n]]
        gaus = amp * np.exp(-(x[None, :]-lamb0[:, None])**2/sigma**2)
        back = bck0
        return np.sum(gaus, axis=0) + back

    def func_sca_jac(x, *args, lamb0=lamb0, n=n):
        amp = np.r_[args[0:n]][None, :]
        sigma = np.r_[args[2*n:3*n]][None, :]
        bck0 = np.r_[args[3*n]]
        lamb0 = lamb0[None, :]
        x = x[:, None]
        jac = np.full((x.size, 2*n+1,), np.nan)
        jac[:, :n] = np.exp(-(x - lamb0)**2/sigma**2)
        jac[:, n:2*n] = amp*2*((x - lamb0)**2/sigma**3
                                 * np.exp(-(x-lamb0)**2/sigma**2))
        jac[:, -1] = 1.
        return jac

    return func_vect, func_sca, func_sca_jac


def multiplegaussianfit1d(x, spectra, nmax=None,
                          lamb0=None, forcelamb=None,
                          p0=None, bounds=None,
                          max_nfev=None, xtol=None, verbose=0,
                          percent=None, plot_debug=False):
    # Check inputs
    if xtol is None:
        xtol = 1.e-8
    if percent is None:
        percent = 20


    # Prepare
    if spectra.ndim == 1:
        spectra = spectra.reshape((1,spectra.size))
    nt = spectra.shape[0]

    # Prepare info
    if verbose is not None:
        print("----- Fitting spectra with {0} gaussians -----".format(nmax))
    nspect = spectra.shape[0]
    nstr = max(nspect//max(int(100/percent), 1), 1)

    # get initial guess function
    if forcelamb is True:
        get_p0bounds = get_p0bounds_lambfix
    else:
        get_p0bounds = get_p0bounds_all

    # lamb0
    if p0 is None or bounds is None or lamb0 is None:
        p00, bounds0, lamb00 = get_p0bounds_all(x, spectra[0,:],
                                              nmax=nmax, lamb0=lamb0)
        if lamb0 is None:
            lamb0 = lamb00
        assert lamb0 is not None
        if forcelamb is True:
            p00 = p00[:nmax] + p00[2*nmax:]
            bounds0 = bounds0[:nmax] + bounds0[2*nmax:]
        if p0 is None:
            p0 = p00
        if bounds is None:
            bounds = bounds0
    if nmax is None:
        nmax = lamb0.size
    assert nmax == lamb0.size

    # Get fit vector, scalar and jacobian functions
    if forcelamb is True:
        func_vect, func_sca, func_sca_jac = get_func1d_lambfix(n=nmax,
                                                               lamb0=lamb0)
    else:
        func_vect, func_sca, func_sca_jac = get_func1d_all(n=nmax,
                                                           lamb0=lamb0)

    # Prepare index for splitting p0
    if forcelamb is True:
        indsplit = nmax*np.r_[1, 2]
    else:
        indsplit = nmax*np.r_[1, 2, 3]

    # Prepare output
    fit = np.full(spectra.shape, np.nan)
    amp = np.full((nt, nmax), np.nan)
    sigma = np.full((nt, nmax), np.nan)
    bck = np.full((nt,), np.nan)
    ampstd = np.full((nt, nmax), np.nan)
    sigmastd = np.full((nt, nmax), np.nan)
    bckstd = np.full((nt,), np.nan)
    if not forcelamb is True:
        dlamb = np.full((nt, nmax), np.nan)
        dlambstd = np.full((nt, nmax), np.nan)
    else:
        dlamb, dlambstd = None, None

    # Loop on spectra
    lch = []
    for ii in range(0, nspect):

        if verbose is not None and ii%nstr==0:
            print("=> spectrum {0} / {1}".format(ii+1, nspect))

        try:
            popt, pcov = scpopt.curve_fit(func_sca, x, spectra[ii,:],
                                          jac=func_sca_jac,
                                          p0=p0, bounds=bounds,
                                          max_nfev=max_nfev, xtol=xtol,
                                          x_scale='jac',
                                          verbose=verbose)
        except Exception as err:
            msg = "    Convergence issue for {0} / {1}\n".format(ii+1, nspect)
            msg += "    => %s\n"%str(err)
            msg += "    => Resetting initial guess and bounds..."
            print(msg)
            try:
                p0, bounds, _ = get_p0bounds(x, spectra[ii,:],
                                             nmax=nmax, lamb0=lamb0)
                popt, pcov = scpopt.curve_fit(func_sca, x, spectra[ii,:],
                                              jac=func_sca_jac,
                                              p0=p0, bounds=bounds,
                                              max_nfev=max_nfev, xtol=xtol,
                                              x_scale='jac',
                                              verbose=verbose)
                p0 = popt
                popt, pcov = scpopt.curve_fit(func_sca, x, spectra[ii,:],
                                              jac=func_sca_jac,
                                              p0=p0, bounds=bounds,
                                              max_nfev=max_nfev, xtol=xtol,
                                              x_scale='jac',
                                              verbose=verbose)
                lch.append(ii)
            except Exception as err:
                print(str(err))
                import ipdb
                ipdb.set_trace()
                raise err

        out = np.split(popt, indsplit)
        outstd = np.split(np.sqrt(np.diag(pcov)), indsplit)
        if forcelamb is True:
            amp[ii, :], sigma[ii, :], bck[ii] = out
            ampstd[ii, :], sigmastd[ii, :], bckstd[ii] = outstd
        else:
            amp[ii, :], dlamb[ii, :], sigma[ii, :], bck[ii] = out
            ampstd[ii,:], dlambstd[ii,:], sigmastd[ii,:], bckstd[ii] = outstd
        fit[ii, :] = func_sca(x, *popt)
        p0[:] = popt[:]

        if plot_debug and ii in [0,1]:
            fit = func_vect(x, amp[ii,:], x0[ii,:], sigma[ii,:], bck0[ii])

            plt.figure()
            ax0 = plt.subplot(2,1,1)
            ax1 = plt.subplot(2,1,2, sharex=ax0, sharey=ax0)
            ax0.plot(x,spectra[ii,:], '.k',
                     x, np.sum(fit, axis=0), '-r')
            ax1.plot(x, fit.T)

    std = np.sqrt(np.sum((spectra-fit)**2, axis=1))

    dout = {'fit': fit, 'lamb0': lamb0, 'std': std, 'lch': lch,
            'amp': amp, 'ampstd': ampstd,
            'sigma': sigma, 'sigmastd': sigmastd,
            'bck': bck, 'bckstd': bckstd,
            'dlamb': dlamb, 'dlambstd': dlambstd}

    return dout



###########################################################
###########################################################
#
#           1d spectral fitting with physics parameters
#
###########################################################
###########################################################


def get_lamb0_from_dlines(dlines):
    lamb0, ions = zip(*[(vv['lamb0'],
                         np.full((len(vv['lamb0']),), kk))
                       for kk, vv in dlines.items()])
    lamb0 = np.r_[lamb0]
    ind = np.argsort(lamb0)
    return lamb0[ind], np.concatenate(ions)[ind]


def get_dindx(bckdeg=None, dlines=None, nbs=None):

    nbck = bckdeg + 1
    if nbs is None:
        # 1d spectral fit
        nbs = 1

    i0 = nbck
    lk = ['sigma', 'dlamb', 'amp', 'ntot', 'nlamb']
    dindx= {'bck': np.r_[:nbck],
            'ions':dict.fromkeys(sorted(dlines.keys())),
            'nbs': nbs}
    for kk in dindx['ions'].keys():
        dindx['ions'][kk] = dict.fromkeys(lk)
        dindx['ions'][kk]['sigma'] = i0 + np.r_[:nbs]
        dindx['ions'][kk]['dlamb'] = i0+nbs + np.r_[:nbs]
        nlamb = len(dlines[kk]['lamb0'])
        dindx['ions'][kk]['amp'] = i0+2*nbs + np.r_[:nlamb*nbs]
        dindx['ions'][kk]['nlamb'] = nlamb
        dindx['ions'][kk]['ntot'] = (2 + nlamb)*nbs
        i0 += dindx['ions'][kk]['ntot']
    dindx['nall'] = i0
    return dindx


def get_x0_bounds(x01d=None, dlines=None, dindx=None,
                  lamb=None, data=None):

    x0 = np.zeros((dindx['nall'],), dtype=float)
    if x01d is None:
        # Get average spectral width and separation
        lamb0_Delta = lamb0.max() - lamb0.min()
        nlamb0 = lamb0.size
        lamb0_delta = lamb0_Delta / nlamb0

        nbs = dindx['nbs']

        x0[dindx['bck']] = np.zeros((dindx['bck'].size,))
        for kk in dindx['ions'].keys():
            # sigma
            x0[dindx[kk]['sigma']] = lamb0_delta
            # dlamb
            x0[dindx[kk]['dlamb']] = 0.
            # amp
            x0[dindx[kk]['amp']] = ampmean

    else:
        x0[dindx['bck']] = x01d[dindx['bck']]
        i0 = dindx['bck'].size
        for kk in dindx['ions'].keys():
            # TBF
            # x0[dindx[kk]['sigma']] = x01d[]
            pass

    # Get bounds
    lamb_delta = np.mean(np.abs(np.diff(np.unique(lamb))))
    datamax = np.nanmax(data)
    bampup = min(datamax, np.nanmean(data) + np.nanstd(data))

    bounds0 = np.zeros((dindx['nall'],), dtype=float)
    bounds1 = np.zeros((dindx['nall'],), dtype=float)
    if dindx['bck'].size == 1:
        bounds0[dindx['bck']] = 0.
        bounds1[dindx['bck']] = bampup
    elif dindx['bck'].size == 2:
        bounds0[dindx['bck'][0]] = 0.
        bounds1[dindx['bck'][0]] = bampup
        bounds0[dindx['bck'][0]] = 0.           # TBC
        bounds1[dindx['bck'][0]] = bampup       # TBC
    for kk in dindx['ions'].keys():
        bounds0[dindx[kk]['sigma']] = 2.*lamb_delta
        bounds1[dindx[kk]['sigma']] = lamb0_delta*5.
        bounds0[dindx[kk]['dlamb']] = -3.*lamb0_delta
        bounds1[dindx[kk]['dlamb']] = 3.*lamb0_delta
        bounds0[dindx[kk]['amp']] = 0.
        bounds1[dindx[kk]['amp']] = datamax

    return x0, bounds




def get_funccostjac():

    def func():
        pass

    def cost():
        pass

    def jac():
        pass

    return func, cost, jac










###########################################################
###########################################################
#
#           2d spectral fitting
#
###########################################################
###########################################################

def get_knots_nbs_for_bsplines(knots_unique, deg):
    if deg > 0:
        knots = np.r_[[knots_unique[0]]*deg, knots_unique,
                      [knots_unique[-1]]*deg]
    else:
        knots = knots_unique
    nbknotsperbs = 2 + deg
    nbs = knots_unique.size - 1 + deg
    assert nbs == knots.size - 1 - deg
    return knots, nbknotsperbs, nbs


def get_2dspectralfit_func(lamb0, forcelamb=False,
                           deg=None, knots=None):

    lamb0 = np.atleast_1d(lamb0).ravel()
    nlamb = lamb0.size
    knots = np.atleast_1d(knots).ravel()
    nknots = knots.size
    nbsplines = np.unique(knots).size - 1 + deg

    # Define function
    def func(lamb, phi,
             camp=None, cwidth=None, cshift=None,
             lamb0=lamb0, nlamb=nlamb,
             knots=knots, deg=deg, forcelamb=forcelamb,
             nbsplines=nbsplines, mesh=True):
        assert phi.ndim in [1, 2]
        if camp is not None:
            assert camp.shape[0] == nbsplines
            bsamp = BSpline(knots, camp, deg,
                            extrapolate=False, axis=0)
        if csigma is not None:
            assert csigma.shape[0] == nbsplines
            bssigma = BSpline(knots, csigma, deg,
                              extrapolate=False, axis=0)
        if mesh or phi.ndim == 2:
            lamb0 = lamb0[None, None, :]
        else:
            lamb0 = lamb0[None, :]
        if forcelamb:
            if mesh:
                assert angle.ndim == lamb.ndim == 1
                # shape (lamb, angle, lines)
                return np.sum(bsamp(phi)[None,:,:]
                              * np.exp(-(lamb[:,None,None]
                                         - lamb0)**2
                                       /(bssigma(phi)[None,:,:]**2)), axis=-1)
            else:
                assert phi.shape == lamb.shape
                lamb = lamb[..., None]
                # shape (lamb/angle, lines)
                return np.sum(bsamp(phi)
                              * np.exp(-(lamb
                                         - lamb0)**2
                                       /(bssigma(phi)**2)), axis=-1)
        else:
            if cdlamb is not None:
                assert cdlamb.shape[0] == nbsplines
                bsdlamb = BSpline(knots, cdlamb, deg,
                                  extrapolate=False, axis=0)

    return func


def get_multigaussianfit2d_costfunc(lamb=None, phi=None, data=None, std=None,
                                    lamb0=None, forcelamb=None,
                                    deg=None, knots=None,
                                    nlamb0=None, nkperbs=None, nbs=None,
                                    nc=None, debug=None):
    assert lamb.shape == phi.shape == data.shape
    assert lamb.ndim == 1
    assert nc == nbs*nlamb0

    if forcelamb is None:
        forcelamb = False
    if debug is None:
        debug = False

    # Define func assuming all inpus properly formatted
    if forcelamb:
        # x = [camp[1-nbs,...,nbs*(nlamb0-1)-nc}, csigma[1-nc]]
        def func(x,
                 lamb=lamb, phi=phi, data=data, std=std,
                 lamb0=lamb0, knots=knots, deg=deg, nc=nc):
            amp = BSpline(knots, x[:nc], deg,
                          extrapolate=False, axis=0)(phi)
            sigma = BSpline(knots, x[nc:], deg,
                            extrapolate=False, axis=0)(phi)
            val = np.sum(amp[:, None]
                         * np.exp(-(lamb[:, None] - lamb0[None, :])**2
                                  /(sigma[:, None]**2)), axis=-1)
            return (val-data)/(std*data.size)

        def jac(x,
                lamb=lamb, phi=phi, std=std,
                lamb0=lamb0, knots=knots, deg=deg,
                nlamb0=nlamb0, nkperbs=nkperbs, nbs=nbs, nc=nc):
            amp = BSpline(knots, x[:nc], deg,
                          extrapolate=False, axis=0)(phi)
            sigma = BSpline(knots, x[nc:], deg,
                            extrapolate=False, axis=0)(phi)
            jacx = sparse.csr_matrix((phi.size, 2*nc), dtype=float)
            #jacx = np.zeros((phi.size, 2*nc), dtype=float)
            for ii in range(nlamb0):
                expi = np.exp(-(lamb-lamb0[ii])**2/sigma**2)
                for jj in range(nbs):
                    ind = ii*nbs + jj
                    indk = np.r_[jj*nkperbs:(jj+1)*nkperbs]
                    # all bsplines are the same, only coefs (x) are changing
                    bj = BSpline.basis_element(knots[indk],
                                               extrapolate=False)(phi)
                    #bj[np.isnan(bj)] = 0.
                    indok = ~np.isnan(bj)
                    # Differentiate wrt camp
                    jacx[indok, ind] = (bj * expi)[indok]
                    # Differentiate wrt csigma
                    jacx[indok, nc+ind] = (
                        amp * (2*(lamb-lamb0[ii])**2*bj/sigma**3) * expi
                    )[indok]
            return jacx/(std*phi.size)
    else:
        # x = [camp1-nbs*nlamb, csigma1-nbs*nlamb, cdlamb1-nbs*nlamb]
        def func(x,
                 lamb=lamb, phi=phi, data=data, std=std,
                 lamb0=lamb0, knots=knots, deg=deg,
                 nbs=nbs, nlamb0=nlamb0, nc=nc, debug=debug):
            amp = BSpline(knots, x[:nc].reshape((nbs, nlamb0), order='F'),
                          deg, extrapolate=False, axis=0)(phi)
            sigma = BSpline(knots, x[nc:2*nc].reshape((nbs, nlamb0), order='F'),
                            deg, extrapolate=False, axis=0)(phi)
            dlamb = BSpline(knots, x[2*nc:-1].reshape((nbs, nlamb0), order='F'),
                            deg, extrapolate=False, axis=0)(phi)
            val = np.nansum(amp
                            * np.exp(-(lamb[:, None] - (lamb0[None, :]+dlamb))**2
                                  / sigma**2),
                            axis=-1) + x[-1]
            if debug:
                vmin, vmax = 0, np.nanmax(data)
                fig = plt.figure(figsize=(14, 10));
                ax0 = fig.add_axes([0.05,0.55,0.25,0.4])
                ax1 = fig.add_axes([0.35,0.55,0.25,0.4], sharex=ax0, sharey=ax0)
                ax2 = fig.add_axes([0.65,0.55,0.25,0.4], sharex=ax0, sharey=ax0)
                ax3 = fig.add_axes([0.05,0.05,0.25,0.4], sharex=ax0, sharey=ax0)
                ax4 = fig.add_axes([0.35,0.05,0.25,0.4], sharex=ax0, sharey=ax0)
                ax5 = fig.add_axes([0.65,0.05,0.25,0.4], sharex=ax0, sharey=ax0)
                ax0.scatter(lamb, phi, c=data, s=2, marker='s', edgecolors='None',
                           vmin=vmin, vmax=vmax)  # DB
                ax1.scatter(lamb, phi, c=val, s=2, marker='s', edgecolors='None',  # DB
                           vmin=vmin, vmax=vmax)  # DB
                errmax = np.nanmax(np.abs((val-data) / (std*data.size)))
                ax2.scatter(lamb, phi, c=(val-data) / (std*data.size),
                            s=2, marker='s', edgecolors='None',  # DB
                            vmin=-errmax, vmax=errmax, cmap=plt.cm.seismic)  # DB
                dlamb0_amp = np.max(np.diff(lamb0))/np.nanmax(np.abs(amp))
                dlamb0_sigma = np.max(np.diff(lamb0))/np.nanmax(np.abs(sigma))
                dlamb0_dlamb = np.max(np.diff(lamb0))/np.nanmax(np.abs(dlamb))
                for ii in range(nlamb0):
                    ax3.axvline(lamb0[ii], ls='--', c='k')
                    ax4.axvline(lamb0[ii], ls='--', c='k')
                    ax5.axvline(lamb0[ii], ls='--', c='k')
                    ax3.plot(lamb0[ii] + dlamb0_amp*amp[:, ii], phi, '.', ms=4)
                    ax4.plot(lamb0[ii] + dlamb0_sigma*sigma[:, ii], phi, '.', ms=4)
                    ax5.plot(lamb0[ii] + dlamb0_dlamb*dlamb[:, ii], phi, '.', ms=4)
                import ipdb         # DB
                ipdb.set_trace()    # DB
            return (val-data) / (std*data.size)

        def jac(x,
                lamb=lamb, phi=phi, std=std,
                lamb0=lamb0, knots=knots, deg=deg,
                nlamb0=nlamb0, nkperbs=nkperbs, nbs=nbs, nc=nc):
            amp = BSpline(knots, x[:nc], deg,
                          extrapolate=False, axis=0)(phi)
            sigma = BSpline(knots, x[nc:2*nc], deg,
                            extrapolate=False, axis=0)(phi)
            dlamb = BSpline(knots, x[2*nc:], deg,
                            extrapolate=False, axis=0)(phi)
            #jacx = sparse.csr_matrix((phi.size, 2*nc), dtype=float)
            jacx = np.zeros((phi.size, 3*nc+1), dtype=float)
            for ii in range(nlamb0):
                expi = np.exp(-(lamb-(lamb0[ii]+dlamb))**2/sigma**2)
                indlamb = expi > 0.001
                for jj in range(nbs):
                    kk = ii*nbs + jj
                    indk = jj + np.r_[:nkperbs]
                    # all bsplines are the same, only coefs (x) are changing
                    bj = BSpline.basis_element(knots[indk],
                                               extrapolate=False)(phi)
                    # bj[np.isnan(bj)] = 0.
                    indok = (~np.isnan(bj)) & indlamb
                    # Differentiate wrt camp
                    jacx[indok, kk] = (bj[indok] * expi[indok])
                    # Differentiate wrt csigma
                    jacx[indok, nc+kk] = (
                        amp * 2*(lamb-(lamb0[ii]+dlamb))**2*bj/sigma**3 * expi
                    )[indok]
                    # Differentiate wrt dlamb
                    jacx[indok, 2*nc+kk] = (
                         amp * 2*(lamb-(lamb0[ii]+dlamb))*bj/sigma**2 * expi
                    )[indok]
            jacx[:, -1] = 1.
            return jacx/(std*phi.size)
    return func, jac

def multigaussianfit2d(lamb, phi, data, std=None,
                       lamb0=None, forcelamb=None,
                       knots=None, deg=None, nbsplines=None,
                       x0=None, bounds=None,
                       method=None, max_nfev=None,
                       xtol=None, ftol=None, gtol=None,
                       loss=None, verbose=0, debug=None):

    # Check inputs
    if deg is None:
        deg = 3
    if nbsplines is None:
        nbsplines = 5
    if method is None:
        method = 'trf'
    # Only 2 method for pb. with bounds
    assert method in ['trf', 'dogbox'], method
    if xtol is None:
        xtol = 1.e-6
    if ftol is None:
        ftol = 1.e-6
    if gtol is None:
        gtol = 1.e-8
    if loss is None:
        loss = 'linear'
    if max_nfev is None:
        max_nfev = None
    if std is None:
        std = 0.1*np.nanmean(data)
    assert lamb0 is not None

    # Get knots
    if knots is None:
        phimin, phimax = np.nanmin(phi), np.nanmax(phi)
        knots = np.linspace(phimin, phimax, nbsplines+1-deg)
    knots, nkperbs, nbs = get_knots_nbs_for_bsplines(np.unique(knots), deg)

    # Scaling
    lambmin = np.nanmin(lamb)
    lamb0Delta = np.max(lamb0) - np.min(lamb0)
    nlamb0 = np.size(lamb0)
    nc = nbs*nlamb0

    dlambscale = lamb0Delta / nlamb0
    ampscale = np.nanmean(data) + np.nanstd(data)

    datascale = data / ampscale
    lambscale = lamb / dlambscale
    lamb0scale = lamb0 / dlambscale
    stdscale = std / ampscale

    # Get cost function and jacobian
    func, jac = get_multigaussianfit2d_costfunc(lamb=lambscale,
                                                phi=phi,
                                                data=datascale,
                                                std=stdscale,
                                                lamb0=lamb0scale,
                                                forcelamb=forcelamb,
                                                deg=deg, knots=knots,
                                                nlamb0=nlamb0, nbs=nbs,
                                                nkperbs=nkperbs, nc=nc,
                                                debug=debug)

    # Get initial guess
    if x0 is None:
        x0 = np.r_[np.ones((nc,)), np.ones((nc,))]
        if not forcelamb:
            x0 = np.r_[x0, np.zeros((nc,))]
            x0 = np.r_[x0, 0.]

    # Get bounds
    if bounds is None:
        bounds = (np.r_[np.zeros((nc,)),
                        np.full((nc,), nlamb0/100)],
                  np.r_[np.full((nc,), np.nanmax(data)/ampscale),
                        np.full((nc,), 3.)])
        if not forcelamb:
            bounds = (np.r_[bounds[0], -np.full((nc,), 2.)],
                      np.r_[bounds[1], np.full((nc,), 2.)])
        bounds = (np.r_[bounds[0], 0.],
                  np.r_[bounds[1], 0.1*np.nanmax(data)/ampscale])

    # Minimize
    res = scpopt.least_squares(func, x0, jac=jac, bounds=bounds,
                               method=method, ftol=ftol, xtol=xtol,
                               gtol=gtol, x_scale=1.0, f_scale=1.0, loss=loss,
                               diff_step=None, tr_solver=None,
                               tr_options={}, jac_sparsity=None,
                               max_nfev=max_nfev, verbose=verbose,
                               args=(), kwargs={})

    # Separate and reshape output
    camp = res.x[:nc].reshape((nlamb0, nbs)) * ampscale
    csigma = res.x[nc:2*nc].reshape((nlamb0, nbs)) * dlambscale
    if forcelamb:
        cdlamb = None
    else:
        cdlamb = res.x[2*nc:3*nc].reshape((nlamb0, nbs)) * dlambscale

    # Create output dict
    dout = {'camp': camp, 'csigma': csigma, 'cdlamb': cdlamb, 'bck':res.x[-1],
            'fit':(func(res.x)*stdscale*data.size + datascale) * ampscale,
            'lamb0':lamb0, 'knots': knots, 'deg':deg, 'nbsplines': nbsplines,
            'cost': res.cost, 'fun': res.fun, 'active_mask': res.active_mask,
            'nfev': res.nfev, 'njev': res.njev, 'status': res.status}
    return dout




###########################################################
#
#           From DataCam2D
#
###########################################################
###########################################################


# DEPRECATED
def fit_spectra2d_x0_per_row():

    # Loop from centre to edges
    for jj in range(ny):
        out = multiplegaussianfit(x, datat[jj,:], nmax=nmax, p0=p0u, bounds=None,
                                  max_nfev=None, xtol=1.e-8, verbose=0,
                                  percent=20, plot_debug=False)
        x0[jj,:], x0_std[jj,:] = out[:2]

        for jj in range(nybis):
            # Upper half
            ju = indy1[jj]
            out = multiplegaussianfit(x, spect, nmax=nmax, p0=p0u, bounds=None,
                                      max_nfev=None, xtol=1.e-8, verbose=0,
                                      percent=20, plot_debug=False)
            x0[ju,:], x0_std[ju,:] = out[:2]
            p0u[:nmax], p0u[nmax:2*nmax], = amp[ii,ju,:], x0[ii,ju,:]
            p0u[2*nmax:3*nmax], p0u[3*nmax:] = sigma[ii,ju,:], bck0

            # Lower half
            jl = indy2[jj]
    return x0



def get_func2d(y0, y1, x0_y, bspl_n, bspl_deg):
    knots = np.linspace(y0,y1, 6)
    bspliney = scpinterp.LSQUnivariateSpline()
    def func(x, y, ampy_coefs, sigy_coefs, bcky_coefs):
        amp_bs = BSpline(knots, ampy_coefs, k=bspl_deg,
                         extrapolate=False, axis=0)
        amp = amp_bs(y)
        x0y = x0_y(y)
        return np.sum(amp*np.exp((x-xoy)**2/sig**2) + bck0, axis=-1)
    return func



def fit_spectra_2d(data2d, indt=None, nbin_init=None,
                   nmax=None, bck=None, nbsplines=None):
    """ Return fitted spectra

    Can handle unique or multiple time
    Takes already formatted 2d data:
        - (nx, ny)
        - (nt, nx, ny)
    x being the horizontal / spectral direction (lambda)

    """

    #####################
    # Check / format input
    #####################

    # Check data
    assert isinstance(data, np.ndarray)
    assert data.ndim in [2,3]
    if data.ndim == 2:
        data = np.reshape((1,data.shape[0],data.shape[1]))
    if indt is not None:
        data = data[indt,...]

    # Set bck type
    if bck is None:
        bck = 0
    assert type(bck) in [int, str]
    if type(bck) is int:
        nbck = bck + 1
    elif bck == 'exp':
        nbck = 3

    # Extract shape
    nt = data.shape[0]
    nlamb, ny = data.shape[1:]
    x = np.arange(0,nlamb)

    # max number of spectral lines (gaussians)
    if nmax is None:
        nmax = 10

    # Check nbin_init vs ny
    if nbin_init is None:
        nbin_init = 100
    if ny % 2 != nbin_init % 2:
        nbin_init += 1

    # get indybin
    indybin = np.arange(0, nbin_init)
    if ny % 2 == 0:
        indybin += int(ny/2 - nbin_init/2)
    else:
        indybin += int((ny-1)/2 - (nbin_init-1)/2)

    # get indybis
    if ny % 2 == 0:
        indy1 = np.arange(ny/2-1, -1, -1)
        indy2 = np.arange(ny/2, ny, 1)
    else:
        indy1 = np.arange((ny-1)/2-1, -1, -1)
        indy2 = np.arange((ny-1)/2+1, ny, 1)
    nybis = indy1.size
    assert nybis == indy2.size

    #####################
    # initialize output
    #####################

    # results
    amp = np.full((nt, ny, nmax), np.nan)
    x0 = np.full((ny, nmax), np.nan)
    sigma = np.full((nt, ny, nmax), np.nan)
    bck0 = np.full((nt, ny, nbck), np.nan)
    p0u = np.full((nmax*3 + nbck,), np.nan)
    p0l = np.full((nmax*3 + nbck,), np.nan)

    # results std
    amp_std = np.full((nt, ny, nmax), np.nan)
    x0_std = np.full((nt, ny, nmax), np.nan)
    sigma_std = np.full((nt, ny, nmax), np.nan)
    bck0_std = np.full((nt, ny, nbck), np.nan)
    std = np.full((nt,), np.nan)


    #####################
    # Determine x0 for each row
    #####################

    datat = np.sum(data, axis=0)

    # Start with central binned (initial guess)

    # if ny is odd => use binned result for central line


    x0 = None

    # Deduce ellipses

    # Derive x0_func(y, x0)


    #####################
    # Define 2d multi-gaussian
    #####################

    # Define x0(y) law




    #####################
    # compute fits, with fixed x0
    #####################

    # Get binned spectra to deduce initial guess
    databin =  None

    # loop over y to fit spectra, start from middle, extent to edges
    for ii in range(nt):

            out = fit2dspectra(x, y, data[ii,:,:])

            p0u[:nmax] = amp[ii-1,indy1[0],:]
            p0u[nmax:2*nmax] = x0[ii-1,indy1[0],:]
            p0u[2*nmax:3*nmax] = sigma[ii,indy1[0],:]
            p0u[3*nmax:] = bck0[ii,indy1[0]]

            p0l[:nmax] = amp[ii-1,indy2[0],:]
            p0l[nmax:2*nmax] = x0[ii-1,indy2[0],:]
            p0l[2*nmax:3*nmax] = sigma[ii,indy2[0],:]
            p0l[3*nmax:] = bck0[ii,indy2[0]]


    return (x0, x0_std), ( )


###########################################################
###########################################################
#
#           For inspiration
#
###########################################################
###########################################################



# _indymod = {0: np.arange(0,195),
            # 1: np.arange(212,407),
            # 2: np.arange(424,619),
            # 3: np.arange(636,831),
            # 4: np.arange(848,1043),
            # 5: np.arange(1060,1255),
            # 6: np.arange(1272,1467)}

# def get_binnedspectra(imp='FeXXV', mod=3, path='./',
                      # dshots=_dshots, indymod=_indymod, save=True):

    # # Prepare
    # indy = indymod[mod]
    # dimp = dict(dshots[imp])

    # # Loop on shots
    # lextra = ['Te0','ne0','dg16']
    # ls = sorted(dimp.keys())
    # for shot in ls:

        # # Load data
        # if 'tlim' in dimp[shot].keys():
            # tlim = dimp[shot]['tlim']
        # else:
            # tlim = None
        # try:
            # xics, kh = tfw.SpectroX2D.load_data(shot,
                                                # tlim=tlim, plot=False)
        # except Exception as err:
            # dimp[shot]['err'] = str(err)
            # continue

        # # Bin data
        # data, t = xics._Ref['data'], xics._Ref['t']
        # dimp[shot]['nt'] = t.size
        # spectra = np.empty((dimp[shot]['nt'], 487), dtype=float)
        # for ii in range(0,dimp[shot]['nt']):
            # spectra[ii,:] = np.nanmean(data[ii,:].reshape(1467,487)[indy,:],
                                       # axis=0)
        # dimp[shot]['spectra'] = spectra
        # dimp[shot]['t'] = t
        # #dimp[shot]['date'] = IRFMtb.

        # # Dextra
        # for ss in lextra:
            # try:
                # indt = np.digitize(xics.dextra[ss]['t'], 0.5*(t[1:]+t[:-1]))
                # val = np.empty((dimp[shot]['nt'],),dtype=float)
                # std = np.empty((dimp[shot]['nt'],),dtype=float)
                # ssum = np.empty((dimp[shot]['nt'],),dtype=float)
                # for ii in range(0,dimp[shot]['nt']):
                    # val[ii] = np.nanmean(xics.dextra[ss]['data'][indt==ii])
                    # std[ii] = np.nanstd(xics.dextra[ss]['data'][indt==ii])
                    # ssum[ii] = np.nansum(xics.dextra[ss]['data'][indt==ii])
                # dimp[shot][ss] = {'mean':val, 'std':std, 'sum':ssum}
            # except Exception as err:
                # dimp[shot][ss] = {'mean':np.nan, 'std':np.nan, 'sum':np.nan}

    # # Reshape dict for np.savez and pandas DataFrame
    # nt = np.array([dimp[shot]['nt'] for shot in ls])
    # nttot = np.sum(nt)
    # ntcum = np.cumsum(nt)
    # lk = ['shot','angle','spectra','t',
          # 'Te0-mean','Te0-std','ne0-mean','ne0-std',
          # 'dg16-sum']
    # dk = {}
    # for k in lk:
        # shape = (nttot,487) if k=='spectra' else (nttot,)
        # dk[k] = np.full(shape, np.nan)

    # i0 = 0
    # for ii in range(0,len(ls)):
        # ind = np.arange(i0,i0+nt[ii])
        # dk['shot'][ind] = ls[ii]
        # dk['angle'][ind] = dimp[ls[ii]]['ang']
        # dk['spectra'][ind,:] = dimp[ls[ii]]['spectra']
        # dk['t'][ind] = dimp[ls[ii]]['t']
        # dk['Te0-mean'][ind] = dimp[ls[ii]]['Te0']['mean']
        # dk['Te0-std'][ind] = dimp[ls[ii]]['Te0']['std']
        # dk['ne0-mean'][ind] = dimp[ls[ii]]['ne0']['mean']
        # dk['ne0-std'][ind] = dimp[ls[ii]]['ne0']['std']
        # dk['dg16-sum'][ind] = dimp[ls[ii]]['dg16']['sum']
        # i0 = ntcum[ii]

    # # Saving
    # if save:
        # name = '%s_spectra'%imp
        # path = os.path.abspath(path)
        # pfe = os.path.join(path,name+'.npz')
        # try:
            # np.savez(pfe, **dk)
            # print("Saved in :", pfe)
        # except:
            # import ipdb
            # ipdb.set_trace()
            # pass
    # return dk



# ####################################################################
# #       spectral fit
# ####################################################################


# def remove_bck(x, y):
    # #opt = np.polyfit(x, y, deg=0)
    # opt = [np.nanmin(y)]
    # ybis = y - opt[0]
    # return ybis, opt[0]


# def get_peaks(x, y, nmax=10):

    # # Prepare
    # ybis = np.copy(y)
    # A = np.empty((nmax,),dtype=y.dtype)
    # x0 = np.empty((nmax,),dtype=x.dtype)
    # sigma = np.empty((nmax,),dtype=y.dtype)
    # gauss = lambda xx, A, x0, sigma: A*np.exp(-(xx-x0)**2/sigma**2)
    # def gauss_jac(xx, A, x0, sigma):
        # jac = np.empty((xx.size,3),dtype=float)
        # jac[:,0] = np.exp(-(xx-x0)**2/sigma**2)
        # jac[:,1] = A*2*(xx-x0)/sigma**2 * np.exp(-(xx-x0)**2/sigma**2)
        # jac[:,2] = A*2*(xx-x0)**2/sigma**3 * np.exp(-(xx-x0)**2/sigma**2)
        # return jac

    # dx = np.nanmin(np.diff(x))

    # # Loop
    # nn = 0
    # while nn<nmax:
        # ind = np.nanargmax(ybis)
        # x00 = x[ind]
        # if np.any(np.diff(ybis[ind:],n=2)>=0.):
            # wp = min(x.size-1,
                     # ind + np.nonzero(np.diff(ybis[ind:],n=2)>=0.)[0][0] + 1)
        # else:
            # wp = ybis.size-1
        # if np.any(np.diff(ybis[:ind+1],n=2)>=0.):
            # wn = max(0, np.nonzero(np.diff(ybis[:ind+1],n=2)>=0.)[0][-1] - 1)
        # else:
            # wn = 0
        # width = x[wp]-x[wn]
        # assert width>0.
        # indl = np.arange(wn,wp+1)
        # sig = np.ones((indl.size,))
        # if (np.abs(np.mean(np.diff(ybis[ind:wp+1])))
            # > np.abs(np.mean(np.diff(ybis[wn:ind+1])))):
            # sig[indl<ind] = 1.5
            # sig[indl>ind] = 0.5
        # else:
            # sig[indl<ind] = 0.5
            # sig[indl>ind] = 1.5
        # p0 = (ybis[ind],x00,width)#,0.)
        # bounds = (np.r_[0.,x[wn],dx/2.],
                  # np.r_[5.*ybis[ind],x[wp],5.*width])
        # try:
            # (Ai, x0i, sigi) = scpopt.curve_fit(gauss, x[indl], ybis[indl],
                                               # p0=p0, bounds=bounds, jac=gauss_jac,
                                               # sigma=sig, x_scale='jac')[0]
        # except Exception as err:
            # print(str(err))
            # import ipdb
            # ipdb.set_trace()
            # pass

        # ybis = ybis - gauss(x, Ai, x0i, sigi)
        # A[nn] = Ai
        # x0[nn] = x0i
        # sigma[nn] = sigi


        # nn += 1
    # return A, x0, sigma

# def get_p0bounds(x, y, nmax=10):

    # yflat, bck = remove_bck(x,y)
    # A, x0, sigma = get_peaks(x, yflat, nmax=nmax)

    # p0 = A.tolist() + x0.tolist() + sigma.tolist() + [bck]

    # lx = [np.nanmin(x), np.nanmax(x)]
    # Dx = np.diff(lx)
    # dx = np.nanmin(np.diff(x))

    # bA = (np.zeros(nmax,), np.full((nmax,),3.*np.nanmax(y)))
    # bx0 = (np.full((nmax,),lx[0]-Dx/2.), np.full((nmax,),lx[1]+Dx/2.))
    # bsigma = (np.full((nmax,),dx/2.), np.full((nmax,),Dx/2.))
    # bbck0 = (0., np.nanmax(y))

    # bounds = (np.r_[bA[0],bx0[0],bsigma[0], bbck0[0]],
              # np.r_[bA[1],bx0[1],bsigma[1], bbck0[1]])
    # if not np.all(bounds[0]<bounds[1]):
        # msg = "Lower bounds must be < upper bounds !\n"
        # msg += "    lower :  %s\n"+str(bounds[0])
        # msg += "    upper :  %s\n"+str(bounds[1])
        # raise Exception(msg)
    # return p0, bounds


# def get_func(n=5):
    # def func_vect(x, A, x0, sigma, bck0):
        # y = np.full((A.size+1, x.size), np.nan)
        # for ii in range(A.size):
            # y[ii,:] = A[ii]*np.exp(-(x-x0[ii])**2/sigma[ii]**2)
        # y[-1,:] = bck0
        # return y

    # def func_sca(x, *args, n=n):
        # A = np.r_[args[0:n]]
        # x0 = np.r_[args[n:2*n]]
        # sigma = np.r_[args[2*n:3*n]]
        # bck0 = np.r_[args[3*n]]
        # gaus = A[:,np.newaxis]*np.exp(-(x[np.newaxis,:]-x0[:,np.newaxis])**2/sigma[:,np.newaxis]**2)
        # back = bck0
        # return np.sum( gaus, axis=0) + back

    # def func_sca_jac(x, *args, n=n):
        # A = np.r_[args[0:n]][np.newaxis,:]
        # x0 = np.r_[args[n:2*n]][np.newaxis,:]
        # sigma = np.r_[args[2*n:3*n]][np.newaxis,:]
        # bck0 = np.r_[args[3*n]]
        # jac = np.full((x.size,3*n+1,), np.nan)
        # jac[:,:n] = np.exp(-(x[:,np.newaxis]-x0)**2/sigma**2)
        # jac[:,n:2*n] = A*2*(x[:,np.newaxis]-x0)/(sigma**2) * np.exp(-(x[:,np.newaxis]-x0)**2/sigma**2)
        # jac[:,2*n:3*n] = A*2*(x[:,np.newaxis]-x0)**2/sigma**3 * np.exp(-(x[:,np.newaxis]-x0)**2/sigma**2)
        # jac[:,-1] = 1.
        # return jac

    # return func_vect, func_sca, func_sca_jac


# def multiplegaussianfit(x, spectra, nmax=10, p0=None, bounds=None,
                        # max_nfev=None, xtol=1.e-8, verbose=0,
                        # percent=20, plot_debug=False):

    # # Prepare
    # if spectra.ndim==1:
        # spectra = spectra.reshape((1,spectra.size))
    # nt = spectra.shape[0]

    # A = np.full((nt,nmax),np.nan)
    # x0 = np.full((nt,nmax),np.nan)
    # sigma = np.full((nt,nmax),np.nan)
    # bck0 = np.full((nt,),np.nan)
    # Astd = np.full((nt,nmax),np.nan)
    # x0std = np.full((nt,nmax),np.nan)
    # sigmastd = np.full((nt,nmax),np.nan)
    # bck0std = np.full((nt,),np.nan)
    # std = np.full((nt,),np.nan)

    # # Prepare info
    # if verbose is not None:
        # print("----- Fitting spectra with {0} gaussians -----".format(nmax))
    # nspect = spectra.shape[0]
    # nstr = max(nspect//max(int(100/percent),1),1)

    # # bounds and initial guess
    # if p0 is None or bounds is None:
        # p00, bounds0 = get_p0bounds(x, spectra[0,:], nmax=nmax)
    # if p0 is None:
        # p0 = p00
    # if bounds is None:
        # bounds = bounds0

    # # Get fit
    # func_vect, func_sca, func_sca_jac = get_func(nmax)
    # lch = []
    # for ii in range(0,nspect):

        # if verbose is not None and ii%nstr==0:
            # print("=> spectrum {0} / {1}".format(ii,nspect))

        # try:
            # popt, pcov = scpopt.curve_fit(func_sca, x, spectra[ii,:],
                                          # jac=func_sca_jac,
                                          # p0=p0, bounds=bounds,
                                          # max_nfev=max_nfev, xtol=xtol,
                                          # x_scale='jac',
                                          # verbose=verbose)
        # except Exception as err:
            # msg = "    Convergence issue for {0} / {1}\n".format(ii,nspect)
            # msg += "    => %s\n"%str(err)
            # msg += "    => Resetting initial guess and bounds..."
            # print(msg)
            # try:
                # p0, bounds = get_p0bounds(x, spectra[ii,:], nmax=nmax)
                # popt, pcov = scpopt.curve_fit(func_sca, x, spectra[ii,:],
                                              # jac=func_sca_jac,
                                              # p0=p0, bounds=bounds,
                                              # max_nfev=max_nfev, xtol=xtol,
                                              # x_scale='jac',
                                              # verbose=verbose)
                # p0 = popt
                # popt, pcov = scpopt.curve_fit(func_sca, x, spectra[ii,:],
                                              # jac=func_sca_jac,
                                              # p0=p0, bounds=bounds,
                                              # max_nfev=max_nfev, xtol=xtol,
                                              # x_scale='jac',
                                              # verbose=verbose)
                # lch.append(ii)
            # except Exception as err:
                # print(str(err))
                # import ipdb
                # ipdb.set_trace()
                # continue


        # A[ii,:] = popt[:nmax]
        # x0[ii,:] = popt[nmax:2*nmax]
        # sigma[ii,:] = popt[2*nmax:3*nmax]
        # bck0[ii] = popt[3*nmax]

        # stdi = np.sqrt(np.diag(pcov))
        # Astd[ii,:] = stdi[:nmax]
        # x0std[ii,:] = stdi[nmax:2*nmax]
        # sigmastd[ii,:] = stdi[2*nmax:3*nmax]
        # bck0std[ii] = stdi[3*nmax]
        # std[ii] = np.sqrt(np.sum((spectra[ii,:]-func_sca(x,*popt))**2))

        # p0[:] = popt[:]

        # if plot_debug and ii in [0,1]:
            # fit = func_vect(x, A[ii,:], x0[ii,:], sigma[ii,:], bck0[ii])

            # plt.figure()
            # ax0 = plt.subplot(2,1,1)
            # ax1 = plt.subplot(2,1,2, sharex=ax0, sharey=ax0)
            # ax0.plot(x,spectra[ii,:], '.k',
                     # x, np.sum(fit,axis=0), '-r')
            # ax1.plot(x, fit.T)

    # return (A,Astd), (x0,x0std), (sigma,sigmastd), (bck0,bck0std), std, lch



# def add_gaussianfits(dimp, nmax=10, verbose=0, percent=20,
                     # path='./', save=False):
    # assert type(dimp) in [str,dict]

    # # Prepare
    # if type(dimp) is str:
        # imp = str(dimp)
        # if '_' in imp:
            # imp = imp.split('_')[0]
        # dimp = dict(np.load(dimp+'_spectra.npz'))
        # inds = np.argsort(dimp['angle'])
        # for k in dimp.keys():
            # if inds.size in dimp[k].shape:
                # dimp[k] = dimp[k][inds] if dimp[k].ndim==1 else dimp[k][inds,:]
    # else:
        # imp = None
        # save = False


    # # Compute
    # ind = np.arange(3,437)
    # x = np.arange(0,dimp['spectra'].shape[1])
    # spectra = dimp['spectra'][:,ind]
    # A, x0, sigma, bck0, std, lch = multiplegaussianfit(x[ind], spectra,
                                                       # nmax=nmax, percent=percent,
                                                       # verbose=verbose)
    # # Store
    # dimp['nmax'] = nmax
    # dimp['indch'] = np.r_[lch]
    # dimp['x'] = x
    # dimp['ind'] = ind
    # dimp['A'] = A[0]
    # dimp['A-std'] = A[1]
    # dimp['x0'] = x0[0]
    # dimp['x0-std'] = x0[1]
    # dimp['sigma'] = sigma[0]
    # dimp['sigma-std'] = sigma[1]
    # dimp['bck0'] = bck0[0]
    # dimp['bck0-std'] = bck0[1]
    # dimp['std'] = std
    # if imp is not None:
        # dimp['imp'] = imp

    # if save:
        # name = '{0}_fitted{1}'.format(imp,nmax)
        # path = os.path.abspath(path)
        # pfe = os.path.join(path,name+'.npz')
        # np.savez(pfe, **dimp)
        # print("Saved in :", pfe)

    # return dimp



# def plot_gaussianfits(dimp, ind):

    # # Prepare
    # x = dimp['x']
    # spectra = dimp['spectra'][ind,:]
    # func_vect, func_sca, func_sca_jac = get_func(dimp['nmax'])
    # fit = func_vect(x, dimp['A'][ind,:], dimp['x0'][ind,:],
                    # dimp['sigma'][ind,:], dimp['bck0'][ind])

    # # Plot
    # plt.figure();
    # ax0 = plt.subplot(2,1,1)
    # ax1 = plt.subplot(2,1,2, sharex=ax0, sharey=ax0)

    # ax0.plot(x, spectra, '.k', label='spectrum')
    # ax0.plot(x, np.sum(fit,axis=0), '-r', label='fit')
    # ax1.plot(x, fit.T)

    # ax1.set_xlabel(r'x')
    # ax0.set_ylabel(r'data')
    # ax1.set_ylabel(r'data')


# def plot_allraw(dimp):

    # # Prepare
    # x = dimp['x']
    # spectra = dimp['spectra']
    # nspect = spectra.shape[0]
    # spectranorm = spectra / np.nanmax(spectra,axis=1)[:,np.newaxis]

    # # Plot
    # plt.figure();
    # ax0 = plt.subplot(2,1,1)
    # ax1 = plt.subplot(2,1,2, sharex=ax0)

    # ax0.imshow(spectranorm, cmap=plt.cm.viridis,
               # extent=(x.min(),x.max(),0,nspect),
               # origin='lower',
               # interpolation='bilinear')


# def extract_lines(dimp):

    # nspect, nx = dimp['spectra'].shape
    # if dimp['imp'] == 'ArXVII':
        # dlines = {'w': {'range':np.arange(280,nx)},
                  # 'z': {'range':np.arange(0,200)}}
        # for k in dlines.keys():
            # spect = np.copy(dimp['spectra'])
            # spect[:,dlines[k]['range']] = 0.
            # dlines[k].update({'x':np.full((nspect,),np.nan),
                              # 'x0':np.full((nspect,),np.nan),
                              # 'A':np.full((nspect,),np.nan),
                              # 'sigma':np.full((nspect,),np.nan)})
            # for ii in range(0,dimp['spectra'].shape[0]):
                # xl = dimp['x'][np.argmax(spect[ii,:])]
                # ind = np.argmin(np.abs(dimp['x0'][ii,:]-xl))
                # dlines[k]['x'][ii] = xl
                # dlines[k]['x0'][ii] = dimp['x0'][ii,ind]
                # dlines[k]['A'][ii] = dimp['A'][ii,ind]
                # dlines[k]['sigma'][ii] = dimp['sigma'][ii,ind]

    # dimp.update(**dlines)
    # return dimp
